#!/usr/bin/python

## @package Backend
# Assistant back-end
# \author Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>
#
# This program is designed to support multiple front-ends to a single
# back-end class.  This makes it easier to write GTK/KDE/command
# line/etc. frontends for the program.  This is the back-end.
#
# @todo Still using protocol version 1

## @file
# Assistant back-end

## This file is distributed under the MIT license
LICENSE = """\
Copyright (c) 2008, Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os, sys, signal, pickle, socket, select, re, gettext, time, subprocess, ConfigParser, tempfile
import two_way_pipe, vnc_interaction, misc, debugger

gettext.install('remote-help-assistant')

## Lowest version of the remote help assistant protocol supported by this version of the program
LOWEST_PROTOCOL_VERSION = 1
## Highest version of the remote help assistant protocol supported by this version of the program
HIGHEST_PROTOCOL_VERSION = 2

## A public key has been altered since the last time it was seen
KEY_CHANGED          = 0
## A public key is in the wrong format
KEY_INVALID          = 1 + KEY_CHANGED
## A public key has never been seen before
KEY_UNKNOWN          = 1 + KEY_INVALID
## A public key needs more permissions before it can be accessed
KEY_NEEDS_PERMISSION = 1 + KEY_UNKNOWN
## A public key is known, and has sufficient permissions for the request
KEY_IS_GOOD          = 1 + KEY_NEEDS_PERMISSION

## Trusted to access another computer
TRUST_ACCESS = 1
## Trusted to share read-only access to the screen
TRUST_SHARE = 2
## Trusted to share read-write access to the keyboard and mouse
TRUST_CONTROL = 4

## Send an HTTP request, receive a response
#
# @param host_addr address of the HTTP server
# @param port port of the HTTP server
# @param end pattern that marks the end of the response
# @param message request to send
# @param timeout number of seconds to wait
#
# @return response, up to and perhaps beyond \c end
#
# This is a fairly simple implementation, designed for the specific
# purpose of UPnP messaging
#
# The \c timeout parameter specifies the maximum time that should be
# taken connecting to the server.  This does not include the time to
# resolve \c host_addr to an IP address.
@misc.log_slow_functions
def get_http_message(host_addr, port, end, message, timeout=4):
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if timeout is not None:
            sock.settimeout(timeout) # Note warning about the timeout 
        sock.connect((host_addr, port))
        sock.setblocking(1)
        sock.sendall(message)
        sock.setblocking(0)
        ewtd = None
        message = ''
        end = re.compile(end)

        while not ewtd:
            iwtd, dummy, ewtd = select.select([ sock ], [], [ sock ])
            if iwtd:
                message += sock.recv(4096)
                if end.search(message):
                    return message
    except (socket.timeout, socket.gaierror, socket.error):
        return None

    return message

## Get information about a UPnP IGD device
#
# @return \c host_addr, \c port, \c host, \c control_url
# 
# This function finds an Internet Gateway with a WANIPConnection
# service.  The WANIPConnection service allows computers on the
# LAN to automatically request connection forwarding.
@misc.log_slow_functions
def get_gateway():
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    message, address = False, False

    # Because UDP is unreliable, this message is sent up to 4 times
    if debug.contact_external_sites():
        for dummy in range(0, 3):

            try:
                # Search for a gateway
                sock.sendto('M-SEARCH * HTTP/1.1\r\n'
                            'HOST: 239.255.255.250:1900\r\n'
                            'MAN: "ssdp:discover"\r\n'
                            'MX: 1\r\n'
                            'ST: urn:schemas-upnp-org:device:InternetGatewayDevice:1\r\n'
                            '\r\n', ('239.255.255.250', 1900))

                iwtd = select.select([ sock ], [], [], 2)[0]
            except (socket.error, select.error):
                return None

            if iwtd:
                message, address = sock.recvfrom(4096)
                break

    # all attempts have failed
    if not address:
        return None

    # Gateway found

    match = re.compile('\nlocation: *http://([^/\r\n]+)([^\r\n]+)', re.IGNORECASE).search(message)
    if match is None:
        debug.log("Unrecognised location in response to M-SEARCH: {" + message + "}")
        return None
    host, page = match.groups()

    match = re.compile('\nUSN: *([^\r\n]+)', re.IGNORECASE).search(message)
    if match is None:
        debug.log("Unrecognised USN in response to M-SEARCH: {" + message + "}")
        return None

    host_addr = host.split(':')
    if 2 == len(host_addr):
        port = int(host_addr[1])
    else:
        port = 80
    host_addr = host_addr[0]

    # See whether the gateway supports the WANIPConnection service

    service_type = re.compile('<serviceType>urn:schemas-upnp-org:service:WANIPConnection:1</serviceType>')
    control_url = None

    services = get_http_message(host_addr, port, "</root>",
                                'GET ' + page + ' HTTP/1.1\r\n'
                                'HOST: ' + host + '\r\n'
                                'Connection: Close\r\n'
                                '\r\n')

    if services is not None:
        for service in services.split('<service>'):
            if service_type.search(service):
                control_url = re.compile('<controlURL>(.*)</controlURL>').search(service).group(1)
                break

    # Doesn't support WANIPConnection
    if control_url is None:
        return None

    return host_addr, port, host, control_url

## Sends a SOAP message and receives a response
# 
# @param action SOAP action to send
# @param upnp_data return value from get_gateway()
# @param args arguments to the SOAP action
# 
# @return contents of the {action}Response element
# 
# SOAP is a protocol used by UPnP routers
def send_soap_message(action, upnp_data, args=None):

    host_addr, port, host, path = upnp_data

    message = \
        '<?xml version="1.0"?>\r\n' \
        '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">\r\n' \
        ' <s:Body>\r\n' \
        '  <u:' + action + ' xmlns:u="urn:schemas-upnp-org:service:WANIPConnection:1">\r\n'

    if args is not None:
        for key, value in args.iteritems():
            message += \
                '   <' + key + '>' + value + '</' + key + '>\r\n'

    message += \
        '  </u:' + action + '>\r\n' \
        ' </s:Body>\r\n' \
        '</s:Envelope>'

    response = \
        get_http_message(host_addr, port, '</s:Envelope>',
                         'POST ' + path + ' HTTP/1.1\r\n'
                         'Connection: Close\r\n'
                         'HOST: ' + host + '\r\n'
                         'CONTENT-LENGTH: ' + str(len(message)) + '\r\n'
                         'CONTENT-TYPE: text/xml; charset="utf-8"\r\n'
                         'SOAPACTION: "urn:schemas-upnp-org:service:WANIPConnection:1#' + action + '"\r\n'
                         '\r\n' + message
                         )

    # Some routers (maybe using an old version of the UPnP standard?) use urn:upnp-org:serviceId:WANIPConnection
    action_namespace = \
        re.compile('xmlns:([^"<> ]*)="(urn:schemas-upnp-org:service:WANIPConnection:1|urn:upnp-org:serviceId:WANIPConnection)"') \
        .search(response)

    if action_namespace is None:

        debug.log("Unparseable response(1) to", action, "\nrequest: {" + message + "}\nresponse: {" + response + "}\n")

        return None, response

    action_namespace = action_namespace.group(1)
    action = action_namespace + ':' + action

    ret = re.compile('<' + action + 'Response[^>]*>(.*?)</' + action + 'Response>', re.DOTALL).search(response)

    if ret is None:

        debug.log("Unparseable response(2) to", action, "\nrequest: {" + message + "}\nresponse: {" + response + "}\n")

        return None, response

    else:
        return ret.group(1), response


## Convert an RSA fingerprint to something nicer
# 
# RSA fingerprints are long and monotonous, and the regular colons
# train me to hear "blah blah colon" and ignore the actual numbers.
# 
# This function converts a fingerprint to a series of signed 32-bit
# integers, then converts those integers to dates.  These dates
# should be acceptable as input to strftime(), although they don't
# necessarily take account of minor date variations.
# 
# This function was written because it's more important for dates to
# be portable than precise, so it's best not to leave things to the
# vagueries of the system's gmtime().
# 
# This function uses midnight on January 1st 1900 as its epoch.
# That date has the nice properties of starting on a Monday and not
# being a leap year even though it's divisible by 4.  These
# properties make the maths a little easier.
@misc.log_slow_functions
def friendly_fingerprint(fingerprint):

    dates = []

    fingerprint = fingerprint.replace(':', '')
    for n in range(0, len(fingerprint) / 8):
        date = int(fingerprint[n*8 : (n+1)*8], 16)

        tm_sec      =    0 + (date % 60) ; date /= 60
        tm_min      =    0 + (date % 60) ; date /= 60
        tm_hour     =    0 + (date % 24) ; date /= 24
        tm_wday     = date % 7 # Dates start from Mon January 1, 1900
        tm_year     = int((date + 1) / 365.25)  ; date -= int(365.25 * tm_year)
        tm_yday     = date + 1
        tm_mday     = tm_yday
        tm_mon      = 1

        if tm_year and not (tm_year % 4) and tm_mday <= 31 + 28: # leap day ahoy!
            if tm_mday < 31:
                tm_mday += 1
            else:
                tm_mon = 2
                tm_mday -= 30
        else:
            for month in (
                31, 28, 31,
                30, 31, 30,
                31, 31, 30,
                31, 30, 31 ):
                if tm_mday <= month:
                    break
                tm_mon += 1
                tm_mday -= month

        #print (tm_year, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, tm_yday, 0)

        # "%c" gives uncomfortably short dates, but it's the only universal alternative that Python gives us.
        # Regional formatting conventions could be supported with translations, but would introduce too much scope for confusion.
        dates.append(time.strftime("%c", (tm_year + 1900, tm_mon, tm_mday, tm_hour, tm_min, tm_sec, tm_wday, tm_yday, 0)))

    return "\n".join(dates)

## A dictionary to convert letters to the NATO phonetic alphabet
IP_ADDRESS_WORDS = {
 '.': _('dot'),
 ':': _('colon'),
 '0': _('zero'),
 '1': _('one'),
 '2': _('two'),
 '3': _('three'),
 '4': _('four'),
 '5': _('five'),
 '6': _('six'),
 '7': _('seven'),
 '8': _('eight'),
 '9': _('nine'),
 'a': 'alpha',
 'b': 'bravo',
 'c': 'charlie',
 'd': 'delta',
 'e': 'echo',
 'f': 'foxtrot',
}


## Translate an IP address to words
#
# When spelling out an IP address over the telephone, it's best to
# pronounce each number individually, so as to reduce the
# possibility of miscommunications.  This function does that.
# 
# Note that for IPv6 addresses, it will use the NATO phonetic
# alphabet letters for A-F.
@misc.log_slow_functions
def spell_IP_address(string):
    if None != re.search('[^A-F0-9.:]', string):
        debug.log('Got impossible IP address,', string)
        return string

    return ' '.join(map(lambda char: IP_ADDRESS_WORDS[char], string))

## Kill specified children on exit
def handle_sigkill(signum, frame):
    debug.log("Killing children: ", Backend.child_pids, trace=True)

    for pid in Backend.child_pids:
        try:
            assert(pid)
            debug.log("Killing", pid)
            os.kill(pid, 15)
        except OSError:
            pass

    while True:
        try:
            os.waitpid(-1, 0)
        except OSError:
            return

## Handle incoming SIGUSR1
# 
# When the assistant is shutting down, it must undo any changes it
# made to the firewall.  This would normally be handled by passing a
# command to the back-end and getting a callback, but the
# command/callback stream might have been closed or repurposed by the
# time the firewall is checked.  The assistant
#
# Instead of using a stream, this message is handled by sending
# signals back and forth: the back-end sends SIGUSR1 to the front-end
# to request that the firewall be reset, and the front-end sends
# SIGUSR1 to the back-end in order to affirm that the firewall can be
# reset.  If the front-end doesn't send an affirmation, the firewall
# is reset when the front-end exits
def handle_sigusr1(signum, frame):
    Backend.__got_sigusr1 = True

## Handle incoming SIGUSR2
#
# When the join_net1_sec2.py is ready, it sends SIGUSR2 to tell the
# backend to go.
sigusr2_seen = False
## Set \c sigusr2_seen to \c True
def handle_sigusr2(signum, frame):
    global sigusr2_seen #pylint: disable-msg=I0011,W0603
    sigusr2_seen = True


## Manages the actual work of the assistant
#
# This class implements the assistant back-end, which connects to the
# other computer, starts a VNC client, and so on.
# 
# The class creates a new process to do the actual work, so it should
# never block the front-end for any length of time.  This class uses a
# simple message-passing system: every command takes a callback
# function, which is called when the command completes (or some other
# command-specific event occurs).  Callbacks are handled by
# poll_message_queue(), which the front-end should run on a regular
# basis.
#
# \section backend_backend_usage Usage
#
# Front-ends will generally use Backend instances by calling each
# command in order: get_network_access(), optionally set_firewall(),
# connect(), start_session(), and finally main_loop().  See the
# individual commands, or \ref backend_protocol "Network protocol" for
# more detail.
#
# \section backend_backend_commands Commands
#
# Most of the member functions of this class are called "back-end
# commands", because they command the back-end to continue the process
# of connecting the assistants.  Back-end commands all have the same
# signature:
# 
# \code
# 	backend.command(argument, callback)
# \endcode
#
# The argument contains the parameters (which might be a list of
# parameters, or might be \c None), and the callback is a function
# that will be called when the command is at a significant point
# (usually when it exits).  Commands that are cancelled with
# cancel_current() probably won't call their callback, although timing
# issues make it impossible to say for sure.
#
# Callback functions must have the signature:
#
# \code
# 	callback(data)
# \endcode
#
# The callback data is specific to each function, and again may be
# any type, including a tuple of values or \c None.  Some functions
# call their callback more than once, some not at all.
# 
# \section backend_backend_implementation Implementation details
#
# A new process is used rather than a new thread, because
# multi-threaded processes are much harder to debug, with no real
# benefit in this case.
#
# \subsection backend_backend_implementation_layers Layered solution
#
# Internally, the backend uses a layered set of protocols.  Other
# implementations are recommended to use the same approach internally.
#
# - Layer 1: network ("net1 layer").  This handles insecure TCP
#   communication between assistants
# - Layer 2: security ("sec2 layer").  This adds security to the
#   previous layer
# - Layer 3: application ("app3 layer").  This handles various
#   application protocols (only RFB is supported at the time of
#   writing)
#
# \sa \ref backend_protocol "Network protocol"
class Backend:

    ## Has SIGUSR1 been received recently?
    # \sa handle_sigusr1()
    __got_sigusr1 = False

    ## If this process is killed, make sure to kill the following child processes...
    child_pids = []

    signal.signal(signal.SIGUSR1, handle_sigusr1)
    signal.signal(15, handle_sigkill)

    ## Constructor
    def __init__(self):

        ## Filename of the current key
        self.__current_key = None

        if __name__ != '__main__': # Called as a library

            from_backend, to_frontend = os.pipe()
            from_frontend, to_backend = os.pipe()

            ## Process ID of the backend process
            self.__pid = os.fork()
            # Fork this file as its own program
            if not self.__pid:
                debug.register_fork()
                os.close(  to_backend)
                os.close(from_backend)

                os.dup2(  to_frontend, sys.stdout.fileno())
                os.dup2(from_frontend, sys.stdin .fileno())

                os.execv(misc.ASSISTANT_PATH + '/backend.py', [ misc.ASSISTANT_PATH + '/backend.py' ] )

            ## Pipe connecting the back-end and front-end
            self.__message_pipe = two_way_pipe.From_filehandles(os.fdopen(from_backend, 'rb'), os.fdopen(to_backend  , 'wb'))

            os.close(  to_frontend)
            os.close(from_frontend)

            ## Next callback ID to use
            self.__next_id = 0
            ## Callback ID of the first value in \c __callbacks
            self.__base_id = 0
            ## List of callback data
            self.__callbacks = []

            ## Configuration data from config.cfg
            self.config = ConfigParser.RawConfigParser()
            self.config.read(misc.CONFIG_DIR + '/config.cfg')

            # Import configuration from pre-0.2 versions of the assistant
            if not self.config.has_section('back-end'):
                self.config.add_section('back-end')
                self.config.set('back-end', 'previous-version', '0.2')
                self.config.set('back-end', 'access-mode', 'no')
                self.config.set('back-end', 'recently-accessed', '')
                if os.path.exists(misc.CONFIG_DIR + "/previous_server") and not self.config.has_option('back-end', 'recently-shared'):
                    self.config.set('back-end', 'recently-shared', open(misc.CONFIG_DIR + "/previous_server", 'r').readline().strip())
                else:
                    self.config.set('back-end', 'recently-shared', '')

            return
        else:

            self.__pid = None

            # Data used by the backend

            ## Parent PID
            self.__ppid = os.getppid()

            self.config = None

            ## Pipe connecting the back-end and front-end
            self.__message_pipe = two_way_pipe.From_stdio()

            signal.signal(signal.SIGUSR1, handle_sigusr1)

            ## IP address of the interface to listen on
            self.__default_interface = None
            ## Was the last call to set_firewall() True or False?
            self.__firewall_enabled = False
            ## cached to avoid calling it more than once
            self.__whatismyip = ''
            ## Information necessary to do UPnP
            self.__upnp_data = get_gateway()
            ## Filename of the current key
            self.__current_key = misc.CONFIG_DIR + "/current_key." + str(os.getpid())

            ## Queue of messages waiting to be processed
            self.__queued_items = [] # May be added to by children

            ## If in sharer mode, should we grant control to the sharer?
            self.__allow_control = False

            ## Username passed to us by the other assistant
            self.__their_username = None
            ## Public key passed to us by the other assistant
            self.__their_public_key = None
            ## username\@host passed to us by the other assistant
            self.__their_account = None


            ## Two-way pipe sending data down into the net1 layer
            self.__net1_top = None
            ## If the other side responds too quickly, connect() might read data intended for start_session()
            self.__net1_buffer = None
            ## Older assistants don't have a __net1_buffer, so wait until the assistant is probably ready
            self.__connect_time = None

            ## two-way pipe for passing data down into the sec2 layer
            self.__sec2_top = None

            ## Two-way pipe for sending data up into the net3 layer
            self.__app3_bottom = None

            ## PID of the process that joins \c __net1_top to \c __sec2_bottom
            self.__join_pid = None


            ## Was there a key associated with this account before the assistant ran?
            self.__had_account = False

            ## Permissions previously associated with this key
            self.__previous_permissions = 0

            ## Is this session in "helper" or "sharer" mode?
            self.__in_accesser_mode = None

            ## Bitmask of permissions required of the other assistant (1=access, 2=share desktop, 4=share control)
            self.__required_permissions = None

            # When the finally block is reached, detect whether we exited the loop cleanly
            completed_ok = False

            try:

                while self.__message_pipe:
                    if self.__queued_items:
                        function, argument, callback_id = self.__queued_items.pop(0)
                    else:
                        try:
                            self.__message_pipe.blocking_reads()
                            function, argument, callback_id = pickle.load(self.__message_pipe.reader())
                        except EOFError, err:
                            debug.log("Backend received EOFError:", err)
                            function = None
                        except IOError, err:
                            debug.log("Backend received IOError:", err)
                            if self.__message_pipe.closed():
                                function = None
                            else:
                                continue

                    if function is None:
                        break

                    debug.log('Running command', callback_id, '==', function, '(', argument, ')')

                    getattr(self, function)(argument, callback_id)

                    debug.log('Finished command', callback_id, '==', function, '(', argument, ')')

                completed_ok = True
            except Exception, e: #pylint: disable-msg=I0011,W0703
                debug.log("Exception:", e) # Without this, sys.exc_info[0] would be None
            finally:
                if not completed_ok:
                    debug.log("Unexpected error:", sys.exc_info()[0], "- stack trace follows:", trace=True)

                self.__respond(-1, None)
                self.__set_firewall(False)
                try: # Catch an error rather than checking whether os.path.exists()
                     # (os.path.exists() could lead to a race condition)
                    if self.__current_key is not None:
                        os.remove(self.__current_key)
                except OSError:
                    pass
                if self.__sec2_top is not None:
                    self.__sec2_top.close()
                    self.__sec2_top = None
                try:
                    if self.__join_pid is not None:
                        assert(self.__join_pid)
                        os.kill(self.__join_pid, 15)
                        os.waitpid(self.__join_pid, 0)
                except OSError:
                    pass

    ## Destructor
    def __del__(self):
        if self.__pid:
            if not self.affirm_set_firewall():
                try:
                    assert(self.__pid)
                    os.kill(self.__pid, 15)
                except:
                    debug.log('failed to kill pid', self.__pid)

        else:
            if self.__message_pipe is not None:
                self.__message_pipe.close()
            self.__message_pipe = None # Disable sending "done" response
            if self.__sec2_top is not None:
                self.__sec2_top.close()
            self.__sec2_top = None
            self.__set_firewall(False)
            handle_sigkill(0, 0)

    ## Update the config.cfg file
    def __save_config(self):
        config_file = open(misc.CONFIG_DIR + '/config.cfg.new', 'wb')
        self.config.write(config_file)
        config_file.flush()
        os.fsync(config_file.fileno())
        config_file.close()
        config_file = None
        os.rename(misc.CONFIG_DIR + '/config.cfg.new', misc.CONFIG_DIR + '/config.cfg')

    ## Get configuration data for a specfified .rha file
    def get_rha_file(self, rha_file):
        try:
            config = ConfigParser.RawConfigParser()
            config.read(os.path.expanduser('~/' + rha_file))

            if not config.has_section('configurator'): return None

            ret = {
                'filename'   : os.path.expanduser('~/' + rha_file),
                }
            has_changes = False

            for choice in ( 'choice-control', 'choice-firewall'):
                if config.has_option('configurator', choice) and (
                    (not self.config.has_option('back-end', choice) ) or
                    config.getboolean('configurator', choice) != self.config.getboolean('back-end', choice) ):
                    ret[choice] = config.get('configurator', choice)
                    has_changes = True

            if config.has_option('configurator', 'choice-mode') and (
                (not self.config.has_option('back-end', 'choice-mode') ) or
                config.get('configurator', 'choice-mode') != self.config.get('back-end', 'choice-mode') ):
                ret['choice-mode'] = config.get('configurator', 'choice-mode')
                has_changes = True

            if config.has_option('configurator', 'choice-ip-editable'):
                if (not self.config.has_option('back-end', 'choice-ip-editable') ) or \
                        config.get('configurator', 'choice-ip-editable') != self.config.get('back-end', 'choice-ip-editable'):
                    ret['choice-ip-editable'] = config.get('configurator', 'choice-ip-editable')
                    has_changes = True
                if (not self.config.get('back-end', 'recently-shared') ) or \
                        self.config.get('back-end', 'recently-shared').split()[0] != config.get('configurator', 'choice-ip-address'):
                    ret['choice-ip-address'] = config.get('configurator', 'choice-ip-address')

            ssh_key = config.get("configurator", "ssh-key")
            their_permissions = config.get("configurator", "ssh-key-permissions").split()
            their_permissions = \
                ( ( 'access'  in their_permissions ) * TRUST_ACCESS ) | \
                ( ( 'share'   in their_permissions ) * TRUST_SHARE  ) | \
                ( ( 'control' in their_permissions ) * TRUST_CONTROL)
            
            their_key, their_account = re.compile(" *([^ ]+) (.*)").match(ssh_key).groups()

            ret['their-account'] = their_account

            if os.path.exists(misc.CONFIG_DIR + '/known_keys'):

                known_keys = open(misc.CONFIG_DIR + '/known_keys')

                for line in known_keys:
                    line = re.compile(" *([^ ]+) (.*?)( [0-7])?\n").match(line)
                    if line:
                        key, account, permissions = line.groups()
                        if account == their_account:
                            if key != their_key or int(permissions) != their_permissions:
                                ret['changed-key'] = key != their_key
                                ret['new-key'] = ssh_key + ' ' + str(their_permissions)
                                ret['new-permissions'] = their_permissions
                                return ret
                            elif has_changes:
                                return ret
                            else:
                                return None

            # Don't add these until we can be certain that the key didn't previously exist
            ret['new-key'] = ssh_key + ' ' + str(their_permissions)
            ret['new-permissions'] = their_permissions

            if has_changes:
                return ret
            else:
                return None
        except:
            debug.log("Unable to read RHA file", rha_file)
            return None

    ## Get data on each Remote Help Configurator file the assistant should process
    def get_rha_files(self):
        return filter(lambda rha: rha is not None,
                      map(self.get_rha_file,
                          filter(lambda path: path.endswith(".rha"),
                                 os.listdir(os.path.expanduser('~')))))

    ## Apply the changes specified in a .rha file
    def apply_rha(self, rha):

        for choice in ( 'choice-mode', 'choice-control', 'choice-firewall', 'choice-ip-editable' ):
            if choice in rha:
                self.config.set('back-end', choice, rha[choice])

        if 'choice-ip-address' in rha:
            address = rha['choice-ip-address']
            recent = self.get_recently_accessed()

            if recent.count(address):
                recent.remove(address)
            recent.insert(0, address)

            self.config.set("back-end", "recently-accessed", " ".join(recent[:5]))

        self.__save_config()

        if 'changed-key' in rha: # Remove all old instances of this key
            new_key = ""
            for line in open(misc.CONFIG_DIR + '/known_keys', 'r').readlines():
                if line.split()[1] != rha['their-account']:
                    new_key += line
                open(misc.CONFIG_DIR + '/known_keys', 'w').write(new_key + rha['new-key'] + "\n")

        elif 'new-key' in rha: # Just append this key
            print >> open(misc.CONFIG_DIR + '/known_keys', 'a'), rha['new-key']

    ## Should the current job terminate before finishing?
    # 
    # @return True if the calling command should quit
    # 
    # Some jobs take a long time to be run, and can be cancelled with
    # cancel_current().  To check whether the job should be
    # cancelled, this function might be called like so:
    #
    # \code
    # 	def my_func:
    # 	while self.__current_job_still_active(1):
    # 	    ... do something ...
    # \endcode

    # @misc.log_slow_functions
    def __current_job_still_active(self, timeout=0):

        # Return True if there's no text waiting to be read
        if not(select.select([ self.__message_pipe ], [], [], timeout)[0]): return True

        self.__message_pipe.nonblocking_reads()
        self.__queued_items.append(pickle.load(self.__message_pipe.reader()))

        return 'cancel_current' != self.__queued_items[-1][0] and self.__queued_items[-1][0] is not None

    ## Send a message to the frontend process
    #
    # @param self Backend object
    # @param callback_id Numeric callback ID passed to the command
    # @param message message to return
    # @param final_call is this the final callback from this command?
    #
    # If \c final_call is \c True, the callback's associated data in the other process
    @misc.log_slow_functions
    def __respond(self, callback_id, message, final_call=True):

        # Don't bother if the frontend has gone away or there's no callback
        if self.__message_pipe and callback_id is not None:
            try:
                debug.log("Responding to", callback_id, '(', message, ')', stack_jump=2)
                # See poll_message_queue() for why this needs to be sent
                self.__message_pipe.write("\n")
                pickle.dump((message, callback_id, final_call), self.__message_pipe.writer(), pickle.HIGHEST_PROTOCOL)
                self.__message_pipe.writer().flush()
            except (IOError, ValueError):
                debug.log("The frontend has gone away")

    ## Send a command to the back-end process
    @misc.log_slow_functions
    def __send_backend_command(self, function, argument, callback):
        try:
            pickle.dump((function, argument, self.__next_id), self.__message_pipe.writer(), pickle.HIGHEST_PROTOCOL)
            self.__message_pipe.writer().flush()
        except IOError:
            debug.log("IO Error while sending command to back-end")
        debug.log("Sent", self.__next_id, '==', function, '(', argument, ')')
        self.__next_id = self.__next_id+1
        self.__callbacks.append(callback)


    ## Check whether the back-end is about to reset the firewall
    #
    # \sa handle_sigusr1()
    @misc.log_slow_functions
    def check_set_firewall(self):
        return Backend.__got_sigusr1

    ## Affirm that the firewall can be reset
    # 
    # It's safe to call this function even when the back-end hasn't
    # requested confirmation
    #
    # \sa handle_sigusr1()
    @misc.log_slow_functions
    def affirm_set_firewall(self):
        if Backend.__got_sigusr1:
            Backend.__got_sigusr1 = False
            os.kill(self.__pid, signal.SIGUSR1)
            return True
        return False

    ## Call any callbacks that are currently waiting
    # 
    # @return \c True if all callbacks were successfully handled, or
    # \c False if a bug caused the back-end to crash.
    # 
    # Front-ends should run this command regularly, in order to
    # receive callbacks from back-end commands.
    # 
    # The frontend should report crashes to the user, ask them to file
    # a report, and quit.
    @misc.log_slow_functions
    def poll_message_queue(self):
        try:
            while 1:
                # If two commands are sent within the time between two
                # calls to poll_message_queue(), select() will only tell you
                # that some data is waiting - there's no way to detect how
                # many messages have been sent.  To get around this
                # problem, we do a non-blocking read for a single
                # (indivisible) character before a blocking read for a
                # pickled object (to avoid issues with half-objects being
                # read)
                self.__message_pipe.nonblocking_reads()
                try:
                    self.__message_pipe.readline()
                except IOError:
                    return True
                self.__message_pipe.blocking_reads()

                message, callback_id, final_call = pickle.load(self.__message_pipe.reader())

                if -1 == callback_id:
                    return False

                self.__callbacks[callback_id - self.__base_id](message)
                if final_call:
                    self.__callbacks[callback_id - self.__base_id] = None

            while self.__callbacks and self.__callbacks[0] is None:
                self.__callbacks.pop(0)
                self.__base_id = self.__base_id+1

        except EOFError:
            debug.log("Got EOF while polling message queue")
            return False

        return True

    ## Has the back-end process died?
    @misc.log_slow_functions
    def has_died(self):
        if self.__pid is None or os.waitpid(self.__pid, os.WNOHANG)[0]:
            self.__pid = None
            return True
        else:
            return False

    ## Get the interface mode (safe or advanced) to use by default
    @misc.log_slow_functions
    def get_default_interface(self):
        if self.config.has_option("back-end", "choice-interface"):
            return self.config.get("back-end", "choice-interface").split('-')[1]
        else:
            return "advanced"

    ## Is the user allowed to change from the default interface?
    @misc.log_slow_functions
    def get_interface_change_allowed(self):
        return ( not self.config.has_option("back-end", "choice-interface") ) or \
            "default" == self.config.get("back-end", "choice-interface").split('-')[0]

    ## Should the assistant default to access interface?
    @misc.log_slow_functions
    def set_default_interface(self, value):
        if value != self.get_default_interface():
            self.config.set("back-end", "choice-interface", 'default-' + 'value')

    ## Get the list of addresses the assistant has recently accessed
    @misc.log_slow_functions
    def get_recently_accessed(self):
        return self.config.get("back-end", "recently-accessed").split()

    ## Get the list of addresses the assistant has recently shared with
    @misc.log_slow_functions
    def get_recently_shared(self):
        return self.config.get("back-end", "recently-shared").split()

    ## Should the assistant default to access mode?
    @misc.log_slow_functions
    def get_access_mode(self):
        return self.config.has_option("back-end", "access-mode") and \
            self.config.getboolean("back-end", "access-mode")

    ## Should the assistant manage the firewall by default?
    @misc.log_slow_functions
    def get_default_firewall(self):
        return self.config.has_option("back-end", "choice-firewall") and \
            self.config.getboolean("back-end", "choice-firewall")

    ## Should the assistant allow control by default?
    @misc.log_slow_functions
    def get_default_control(self):
        return self.config.has_option("back-end", "choice-control") and \
            self.config.getboolean("back-end", "choice-control")

    ## Should the assistant allow the IP address to be altered?
    @misc.log_slow_functions
    def get_default_ip_editable(self):
        return not self.config.has_option("back-end", "choice-ip-editable") or \
            self.config.getboolean("back-end", "choice-ip-editable")


    ## Cancel the currently-running command (if any)
    # 
    # Commands that run for a long time can be cancelled mid-run.  This
    # command signals that the currently running command should exit at
    # its earliest convenience.
    @misc.log_slow_functions
    def cancel_current(self, callback_id=None, callback=None):
        if self.__pid:
            return self.__send_backend_command('cancel_current', None, callback_id)

    ## Information about this computer's network access
    #
    # Before the computers can connect to each other, they must assess
    # what sort of a connection they're capable of making.  For
    # example, is either computer capable of listening on a TCP port
    # on the public Internet?  This command assesses the situation
    # this computer finds itself in.
    # 
    # The argument to this \ref backend_backend_commands "command" is
    # ignored.
    #
    # Once the command has finished, the callback will be called with
    # an argument of the form (\c success\c , \c message\c , \c
    # IP_pronunciation\c , \c router_URL\c , \c private_address)).
    # The elements have the following meanings:
    #
    # - \c success is \c True if the user's information was found
    #   successfully, or \c False otherwise.
    # - \c message is a string for the user, with a meaning depending
    #   on the \c success value.  If \c success is \c True, \c message
    #   is the IP address that should be read out to the other user.
    #   Otherwise, it is an error message.
    # - \c IP_pronunciation is a string describing how to pronounce
    #   the above IP address (or the empty string if \c success is
    #   false).  An example value would be "one dot two dot three dot
    #   four".
    # - \c router_URL is the address of this computer's non-UPnP
    #   router (or \c None if not applicable).  This is provided if
    #   the user should modify settings on a router, and the router
    #   appears to have a web-based interface.  An example value would
    #   be "http://192.168.0.1/"
    # - \c private_address is a string describing the IP address of
    #   this computer on its private network (or an empty string if
    #   this is inapplicable)
    @misc.log_slow_functions
    def get_network_access(self, argument, callback_id, final_call=True):

        if self.__pid:
            return self.__send_backend_command('get_network_access', argument, callback_id)

        # Get our IP address and default interface

        cmd_pipe = os.popen('/sbin/route -n')
        pattern = re.compile('0\.0\.0\.0         ([0-9.]*).* ([a-z0-9]*)\n')
        default_route = self.__default_interface = None

        for line in cmd_pipe:
            route = pattern.match(line)
            if route:
                default_route = route.group(1)
                self.__default_interface = route.group(2)
                break

        if self.__default_interface is None:
            return self.__respond(callback_id,
                                [
                    False,
                    _("This computer doesn't seem to be connected to the Internet. Please rescan after going online."),
                    "", "", ""
                    ],
                                final_call)

        cmd_pipe = os.popen('/sbin/ifconfig ' + self.__default_interface)
        pattern = re.compile('          [^:]*:([0-9.]*)')
        self.__default_interface = None

        for line in cmd_pipe:
            addr = pattern.match(line)
            if (addr):
                self.__default_interface = addr.group(1)
                break

        if self.__default_interface is None:
            return self.__respond(callback_id,
                                  [
                    False,
                    _("An error in this program meant that the assistant couldn't find your IP address. Please report this bug."),
                    "", "", ""
                    ],
                                final_call)

        # If the default gateway or the UPnP server are listening on port 80, assume it's a router's web interface

        router_URL = None

        if default_route:
            if get_http_message(default_route, 80, "^",
                                "GET / HTTP/1.1\r\n"
                                "\r\n") is not None:
                router_URL = "http://" + default_route + "/"

        if self.__upnp_data and not router_URL:
            if get_http_message(self.__upnp_data[0], 80, "^",
                                "GET / HTTP/1.1\r\n"
                                "\r\n") is not None:
                router_URL = "http://" + self.__upnp_data[0] + "/"


        # If a UPnP router was detected, use that

        if self.__upnp_data:
            external_IP = send_soap_message('GetExternalIPAddress', self.__upnp_data)[0]
            if external_IP is not None:
                external_IP = re.compile("<NewExternalIPAddress>\s*(.*?)\s*</NewExternalIPAddress>").search(external_IP)
                if external_IP is not None:
                    external_IP = external_IP.group(1)

                    return self.__respond(callback_id,
                                          [ True, external_IP, spell_IP_address(external_IP), router_URL, self.__default_interface ],
                                          final_call)

        # If this is a public IP address, use that

        pattern = re.compile("""
                  10\.                         |
                  192\.                        |
                  172\.(1[6-9]|2[0-9]|3[01])\. |
                  169.254\.
        """, re.VERBOSE)

        if not pattern.match(self.__default_interface):
            return self.__respond(callback_id,
                                  [ True, self.__default_interface, spell_IP_address(self.__default_interface), router_URL, self.__default_interface ],
                                  final_call)

        # Private, non-UPnP address.

        # Get a probable IP address from whatismyip.com
        if debug.contact_external_sites() and '' == self.__whatismyip:
            myip = get_http_message('whatismyip.com', 80,
                                    '\n\r?\n([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})$',
                                    "GET /automation/n09230945.asp HTTP/1.1\r\nHost: whatismyip.com\r\nConnection: Close\r\n\r\n",
                                    4)

            if myip is None:
                myip = ""
            else:
                myip = re.compile('\n\r?\n([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})$').search(myip).group(1)

            return self.__respond(callback_id,
                                  [ True, myip, spell_IP_address(myip), router_URL, self.__default_interface ],
                                  final_call)

        return self.__respond(callback_id,
                              [ True, "unknown", "couldn't detect your IP address", router_URL, self.__default_interface ],
                              final_call)

    ## Disable/enable firewalls, including any UPnP gateway firewalls
    #
    # @param self Backend instance
    # @param enable \c True to enable firewall management, \c False to
    # disable firewall management.
    # @param callback_id unique ID of the callback
    # @return \c True on success, \c False on failure (e.g. password not entered), or \c None on error
    #
    # This function adds firewall rules for contacting the UPnP
    # gateway device, so it might try to find the gateway device by
    # calling get_network_access(), which will in turn call the
    # callback.
    #
    # If \c callback_id is \c None, the program is assumed to be
    # closing down.  The function therefore waits for the user to
    # affirm changing the firewall
    @misc.log_slow_functions
    def __set_firewall(self, enable, callback_id=None):

        if self.__firewall_enabled == enable:
            return True

        self.__firewall_enabled = enable

        if callback_id is None:
            try:
                os.kill(self.__ppid, signal.SIGUSR1)
                while not Backend.__got_sigusr1 and not os.kill(os.getppid(), 0): # Until the parent confirms or terminates
                    time.sleep(0.5)
                Backend.__got_sigusr1 = False
            except:
                pass

        tmpfile = None

        if   os.path.exists('/usr/bin/gksudo'): command = '/usr/bin/gksudo --description "%s" -- ' % _('Remote help assistant (firewall)')
        elif os.path.exists('/usr/bin/kdesu' ): command = '/usr/bin/kdesu  --comment "%s" -- ' % _('Remote help assistant (firewall)')
        else:
            command = '/usr/bin/xterm -title "%s" -e sudo ' % _('Remote help assistant (firewall)')
            tmpfile = tempfile.NamedTemporaryFile(mode='rwb') # xterm eats standard output

        command += '"' + misc.ASSISTANT_PATH + '/fix_iptables.py' + '"'

        if enable:
            command += ' enable'
        else:
            command += ' disable'

            if self.__upnp_data:
                # Delete any previous port mappings, in case the firewall blocks this later
                send_soap_message('DeletePortMapping', self.__upnp_data,
                                  { 'NewRemoteHost': '',
                                    'NewExternalPort': str(misc.REMOTE_PORT),
                                    'NewProtocol': 'TCP'}
                                  )
                send_soap_message('DeletePortMapping', self.__upnp_data,
                                  { 'NewRemoteHost': '',
                                    'NewExternalPort': str(misc.REMOTE_PORT),
                                    'NewProtocol': 'UDP'}
                                  )

        if tmpfile is None:
            result = os.popen(command).readlines()
        else:
            os.popen(command + ' >\'' + tmpfile.name + '\'').readlines() # Wait for the command to exit
            result = tmpfile.readlines()

        if not result or 0 == len(result):
            return False

        if 'Success\n' != result[0]:
            debug.log("iptables error")
            if tmpfile is not None:
                debug.log("name:", tmpfile.name)

            return None

        # Don't bother with the (slow) remainder of the function if we're about to exit
        if callback_id is None:
            return

        # Handling the backend includes opening up ports needed by UPnP
        self.__upnp_data = get_gateway()
        self.get_network_access(None, callback_id, False)

        if self.__upnp_data:
            # Delete any previous port mappings
            send_soap_message('DeletePortMapping', self.__upnp_data,
                              { 'NewRemoteHost': '',
                                'NewExternalPort': str(misc.REMOTE_PORT),
                                'NewProtocol': 'TCP'}
                              )
            send_soap_message('DeletePortMapping', self.__upnp_data,
                              { 'NewRemoteHost': '',
                                'NewExternalPort': str(misc.REMOTE_PORT),
                                'NewProtocol': 'UDP'}
                              )

            if enable:
                mapping = { 'NewRemoteHost': '',
                            'NewExternalPort': str(misc.REMOTE_PORT),
                            'NewProtocol': 'TCP',
                            'NewInternalPort': str(misc.REMOTE_PORT),
                            'NewInternalClient': self.__default_interface,
                            'NewEnabled': '1',
                            'NewPortMappingDescription': 'Remote Help Assistant',
                            'NewLeaseDuration': '0'}

                if send_soap_message('AddPortMapping', self.__upnp_data, mapping)[0] is None:
                    return None

                mapping['NewProtocol'] = 'UDP'
                if send_soap_message('AddPortMapping', self.__upnp_data, mapping)[0] is None:
                    return None

        return True


    ## Enable or disable the firewall
    # This function is used internally by connect()
    def __manage_firewall(self, manage_firewalls, callback_id, enable):
        if manage_firewalls:
            self.__respond(callback_id, (None, _("Setting up firewalls...")), False)
            ret = self.__set_firewall(enable, callback_id)
            if ret is None:
                self.__respond(callback_id,
                               (1, _("Failed: couldn't manage the firewall"),
                                _("""The assistant can't manage the firewall, because of a bug in the assistant.
Please report this bug.""")))
                return True
            elif ret is False:
                self.__respond(callback_id,
                               (1, _("Failed: couldn't manage the firewall"),
                                _("""The assistant can't manage the firewall, because you do not have permission to modify it.""")))
                return True

        return False

    ## Does \c sock have data waiting to be read?
    def __waiting_data(self, sock):
        try:
            sock.settimeout(1)
            socket_string = sock.recv(1)
            sock.setblocking(0)
            return socket_string
        except socket.error:
            return ''

    ## Listen for incoming TCP connections
    # @note these go before connect_sockets() to ensure that an IPv4 "connecting" callback_msg trumps an IPv6 "listeninng" message
    def __create_listen_socket(self, readers, family, address):
        listen_sock = misc.listen_sockets(address, misc.REMOTE_PORT, family=family)[0]
        if listen_sock is not None:
            readers.append(listen_sock)
        return listen_sock

    ## Create a socket for an outgoing connection
    def __create_connect_socket(self, writers, family, addrinfo):
        tcp_socket = socket.socket(family, socket.SOCK_STREAM)
        tcp_socket.setblocking(0)
        try:
            tcp_socket.connect(addrinfo)
        except socket.error, err:
            if 115 != err[0]: raise err # 115 == Operation now in progress

        writers.append(tcp_socket)

        return tcp_socket

    ## Open a set of sockets for either IPv4 or IPv6
    #
    # These are TCP (connecting) and UDP
    def __connect_sockets(self, family, callback_msg, remote_address, writers):

        # DO NOT PUT listen_sock stuff here (see message above)

        if "" == remote_address:
            return None, None, None, None, None, None, callback_msg

        try: # TCP and UDP
            addrinfo = socket.getaddrinfo(remote_address, misc.REMOTE_PORT, family)

            if len(addrinfo):

                addrinfo = addrinfo[0][4]

                callback_msg = _("Connecting to %s...") % remote_address

                tcp_socket = self.__create_connect_socket(writers, family, addrinfo)

                # UDP
                if os.access("/usr/sbin/pppd", os.X_OK):

                    if socket.AF_INET == family:
                        tunneled_from = '10.152.4.124' ; tunneled_to = '10.152.4.151'
                    else:
                        tunneled_from = '10.152.6.124' ; tunneled_to = '10.152.6.151'

                    if self.__in_accesser_mode:
                        tunneled_from, tunneled_to = tunneled_to, tunneled_from

                    pppd = os.fork()

                    if pppd:
                        if self.__in_accesser_mode:
                            ppp_addrinfo = (tunneled_to, misc.REMOTE_PORT)
                            ppp_socket = self.__create_connect_socket(writers, family, ppp_addrinfo)
                        else:
                            ppp_addrinfo = None
                            ppp_socket = None

                        Backend.child_pids.append(pppd)

                    else:
                        debug.register_fork()
                        cmd = ( misc.ASSISTANT_PATH + '/pppd.py', str(family), remote_address, tunneled_from, tunneled_to )
                        os.execv(cmd[0], cmd)

                else:
                    pppd = ppp_socket = ppp_addrinfo = None

                # Fork a `ping` process rather than doing it ourselves, because you have to be setuid root to ping things
                ping_pid = os.fork()
                if not ping_pid:
                    debug.register_fork()
                    os.close(1)
                    if socket.AF_INET == family:
                        os.execv('/bin/ping' , ( '/bin/ping' , '-q', '-c', '1', '-w', '1000000', remote_address) )
                    else:
                        os.execv('/bin/ping6', ( '/bin/ping6', '-q', '-c', '1', '-w', '1000000', remote_address) )

                Backend.child_pids.append(ping_pid)


            else:
                addrinfo = tcp_socket = pppd = ppp_socket = ppp_addrinfo = None

        except socket.error, error_message:
            if -9 != error_message[0]: # -9 == Address family for hostname not supported
                callback_msg = _("Connecting... %s") % error_message[1]
            addrinfo = tcp_socket = pppd = ppp_socket = ppp_addrinfo = ping_pid = None

        return tcp_socket, addrinfo, pppd, ppp_socket, ppp_addrinfo, ping_pid, callback_msg

    ## Kill children used during connecting
    def __kill_intermediates(self, ppp4_pid, ppp4_sock, ppp6_pid, ppp6_sock, ping4_pid, ping6_pid): #pylint: disable-msg=I0011,R0913
        to_kill = []
        if  ppp4_pid is not None and ppp4_sock != self.__net1_top: to_kill.append( ppp4_pid)
        if  ppp6_pid is not None and ppp6_sock != self.__net1_top: to_kill.append( ppp6_pid)
        if ping4_pid is not None                                 : to_kill.append(ping4_pid)
        if ping6_pid is not None                                 : to_kill.append(ping6_pid)

        for pid in to_kill:
            os.kill(pid, 15)

        for pid in to_kill:
            os.waitpid(pid, 0)
            Backend.child_pids.remove(pid)


    ## Connect to the other assistant
    #
    # This function runs the first stage in the session: creating an
    # insecure connection.  The argument to this \ref
    # backend_backend_commands "command" should be a tuple:
    #
    # \code
    # (
    #  in_accesser_mode, # Is this computer the helper?
    #  remote_address,   # What address should we connect to?
    #  allow_control,    # Should we give them control? (if not helper)
    #  manage_firewalls, # Should we manage firewalls?
    # )
    # \endcode
    #
    # If this computer should not attempt to connect to the other, \c
    # remote_address must be an empty string.
    #
    # The callback will be called at least once, and the callback's \c
    # data will be a tuple containing between two and nine elements:
    #
    # - ( \c fraction\c ,\c message \c ) OR
    # - ( \c fraction\c ,\c message\c ,\c long_message \c )
    # - ( \c success\c  ,\c message\c ,\c router_URL\c , \c IP_pronunciation\c ,\c private_address \c )
    # - ( \c fraction\c ,\c message\c ,\c our_key\c ,\c their_key\c ,\c their_account\c ,\c their_key_trust\c ,\c required_key_trust\c ,\c key_status\c ,\c their_key_status \c )
    #
    # The five-element tuple is actually a callback from the
    # get_network_access() command, which is sometimes run as part of
    # this command.  Aside from this, all tuples have a similar
    # format:
    #
    # The tuple begins with the \c fraction - a value between 0 and 1,
    # specifying how complete the operation is (or None if the
    # function is still trying to connect).  The \c message is a
    # string to show the user explaining what the program is doing (or
    # None to keep the old message).
    #
    # If \c data has three elements, the connection has failed, and the
    # \c long message explains what the problem is.
    #
    # If \c data has nine elements, the connection has succeeded.  \c
    # our_key is our SSH key, represented as a series of dates.  \c
    # their_key is their SSH key as dates.  \c their_account is the
    # account name associated with their key.  \c their_key_trust and
    # \c required_key_trust are bitmasks of \c TRUST_* constants
    # representing the actions the specified user is currently, and is
    # required to, be trusted with.  \c key_status is a long string
    # describing \c their_key.  \c their_key_trusted is one of the \c
    # KEY_* constants.
    #
    # Because the above values have been sent over an insecure
    # connection, they might have been modified in a
    # man-in-the-middle attack.  It's important that the users
    # confer with each other over a tamper-proof line (e.g. a
    # telephone) in order to verify the keys.
    @misc.log_slow_functions
    def connect(self, argument, callback_id):

        if self.__pid:

            self.__send_backend_command('connect', argument, callback_id)

            # Update config settings after sending the command
            self.config.set("back-end", "access-mode", argument[0])

            if argument[0]:
                recent = self.get_recently_accessed()
            else:
                recent = self.get_recently_shared()

            if recent.count(argument[1]):
                recent.remove(argument[1])
            recent.insert(0, argument[1])
            recent = " ".join(recent[:5])

            if argument[0]:
                self.config.set("back-end", "recently-accessed", recent)
            else:
                self.config.set("back-end", "recently-shared", recent)

            return

        self.__in_accesser_mode, remote_address, self.__allow_control, manage_firewalls = argument

        if self.__in_accesser_mode:
            self.__required_permissions = 1
        elif self.__allow_control:
            self.__required_permissions = 3
        else:
            self.__required_permissions = 2

        if self.__manage_firewall(manage_firewalls, callback_id, True):
            self.__respond(callback_id, (1, _("Failed: couldn't set up firewalls")))
            return

        #
        # PART ONE: initialise things
        #

        # Information about the other computer
        socket_string = ''

        # List for use with select()
        readers, writers = [], []

        # foo_unaccepted_listener is used in a dual test of (is this foo_listen_sock?) and (has foo_listen_sock.accept() been called?)
        tcp4_listen_sock = tcp4_unaccepted_listener = self.__create_listen_socket(readers, socket.AF_INET , '0.0.0.0')
        tcp6_listen_sock = tcp6_unaccepted_listener = self.__create_listen_socket(readers, socket.AF_INET6, '::'     )

        # Message to send to the front-end
        if tcp4_listen_sock is None and tcp6_listen_sock is None:
            callback_msg = _("Please wait...")
        else:
            callback_msg = _("Listening for connections...")

        tcp4_conn_sock, tcp4_addr, ppp4_pid, ppp4_sock, ppp4_addr, ping4_pid, callback_msg = \
            self.__connect_sockets(socket.AF_INET , callback_msg, remote_address, writers)
        tcp6_conn_sock, tcp6_addr, ppp6_pid, ppp6_sock, ppp6_addr, ping6_pid, callback_msg = \
            self.__connect_sockets(socket.AF_INET6, callback_msg, remote_address, writers)

        self.__respond(callback_id, (None, callback_msg), False)
        old_msg = callback_msg

        #
        # PART TWO: try to make a connection
        #

        choice_made = False
        reconnecters = []

        while not choice_made:

            iwtd, owtd = select.select(readers, writers, [], 1)[0:2]

            if not self.__current_job_still_active():
                self.__kill_intermediates(ppp4_pid, ppp4_sock, ppp6_pid, ppp6_sock, ping4_pid, ping6_pid)
                self.__net1_top = None
                return self.__manage_firewall(manage_firewalls, callback_id, False)

            if iwtd or owtd:

                for reader in iwtd:
                    # In helper mode, immediately accept connections
                    if tcp4_listen_sock == reader or tcp6_listen_sock == reader:
                        if self.__in_accesser_mode:
                            maybe_reader, address = reader.accept()
                            # Ignore connections from the wrong address
                            if "" == remote_address or address == remote_address:
                                self.__net1_top = two_way_pipe.From_connected_socket(maybe_reader)
                                choice_made = True
                                if   tcp4_listen_sock == reader: debug.log("made an inbound IPv4 TCP connection")
                                elif tcp6_listen_sock == reader: debug.log("made an inbound IPv6 TCP connection")
                                else                           : debug.log("made an unknown connection (1)")

                            break
                        else:
                            if reader == tcp4_unaccepted_listener:
                                maybe_reader, address = reader.accept()
                                # Ignore connections from the wrong address
                                if "" == remote_address or address == remote_address:
                                    readers.remove(reader)
                                    reader = tcp4_listen_sock = maybe_reader
                                    readers.append(reader)
                                    tcp4_unaccepted_listener = None
                                else:
                                    reader = None
                            elif reader == tcp6_unaccepted_listener:
                                maybe_reader, address = reader.accept()
                                # Ignore connections from the wrong address
                                if "" == remote_address or address == remote_address:
                                    readers.remove(reader)
                                    reader = tcp6_listen_sock = maybe_reader
                                    readers.append(reader)
                                    tcp6_unaccepted_listener = None
                                else:
                                    reader = None


                                readers.remove(reader)
                                reader = tcp6_listen_sock = reader.accept()[0]
                                readers.append(reader)
                                tcp6_unaccepted_listener = None

                    if reader is not None:
                        socket_string = self.__waiting_data(reader)
                        if len(socket_string):
                            self.__net1_top = two_way_pipe.From_connected_socket(reader)

                            if   tcp4_conn_sock   == reader: debug.log("made an outbound IPv4 TCP connection")
                            elif tcp6_conn_sock   == reader: debug.log("made an outbound IPv6 TCP connection")
                            elif ppp4_sock        == reader: debug.log("made an outbound IPv4 PPP connection")
                            elif ppp6_sock        == reader: debug.log("made an outbound IPv6 PPP connection")
                            elif tcp4_listen_sock == reader: debug.log("made an inbound IPv4 TCP connection")
                            elif tcp6_listen_sock == reader: debug.log("made an inbound IPv6 TCP connection")
                            else                           : debug.log("made an unknown connection (2)")

                            choice_made = True
                            break

                for writer in owtd:
                    if 0 == writer.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR):
                        # It would be better to do "self.__in_accesser_mode or waiting_data(reader)", but there's no way to make that work in Python
                        choice_made = self.__in_accesser_mode
                        if not choice_made:
                            socket_string = self.__waiting_data(writer)
                            choice_made = len(socket_string)

                        if choice_made:
                            self.__net1_top = two_way_pipe.From_connected_socket(writer)
                            if   tcp4_conn_sock == writer: debug.log("made an outbound IPv4 TCP connection")
                            elif tcp6_conn_sock == writer: debug.log("made an outbound IPv6 TCP connection")
                            elif ppp4_sock      == writer: debug.log("made an outbound IPv4 PPP connection")
                            elif ppp6_sock      == writer: debug.log("made an outbound IPv6 PPP connection")
                            else                         : debug.log("made an unknown connection (3)")
                            break

                        else:
                            writers.remove(writer)
                            readers.append(writer)

                        # Connection succeeded
                    else:
                        reconnecters.append(writer)
                        writers.remove(writer)

            else:

                if None == tcp4_listen_sock: tcp4_listen_sock = tcp4_unaccepted_listener = self.__create_listen_socket(readers, socket.AF_INET , '0.0.0.0')
                if None == tcp6_listen_sock: tcp6_listen_sock = tcp6_unaccepted_listener = self.__create_listen_socket(readers, socket.AF_INET6, '::')

                if "" == remote_address and ( tcp4_listen_sock is not None or tcp6_listen_sock is not None ):
                    callback_msg = _("Listening for connections...")

                for reconnecter in reconnecters:
                    if   tcp4_conn_sock == reconnecter: tcp4_conn_sock = self.__create_connect_socket(writers, socket.AF_INET , tcp4_addr)
                    elif tcp6_conn_sock == reconnecter: tcp6_conn_sock = self.__create_connect_socket(writers, socket.AF_INET6, tcp6_addr)
                    elif ppp4_sock      == reconnecter: ppp4_sock      = self.__create_connect_socket(writers, socket.AF_INET , ppp4_addr)
                    elif ppp6_sock      == reconnecter: ppp4_sock      = self.__create_connect_socket(writers, socket.AF_INET6, ppp6_addr)
                    
                reconnecters = []

                if None != ping4_pid:
                    ret = os.waitpid(ping4_pid, os.WNOHANG)
                    if ret[0]:
                        if not ret[1]:
                            callback_msg = _("Connecting... computer found, assistant not listening")
                    ping4_pid = None

                if None != ping6_pid:
                    ret = os.waitpid(ping6_pid, os.WNOHANG)
                    if ret[0]:
                        if not ret[1]:
                            callback_msg = _("Connecting... computer found, assistant not listening")
                    ping6_pid = None

            # Note: the following tends to get called twice in quick succession when iwtd is non-empty - this is perfectly normal
            if callback_msg != old_msg:
                self.__respond(callback_id, (None, callback_msg), False)
                old_msg = callback_msg

        self.__kill_intermediates(ppp4_pid, ppp4_sock, ppp6_pid, ppp6_sock, ping4_pid, ping6_pid)

        return self.__negotiate(callback_id, socket_string)


    ## Negotiate protocol versions, etc.
    # This was split from connect() to emphasise their seperate purposes - there's no functional need for it
    def __negotiate(self, callback_id, socket_string):

        debug.log(
            "- Sending the following information:",
            str(LOWEST_PROTOCOL_VERSION) + str(HIGHEST_PROTOCOL_VERSION) +
            str(self.__in_accesser_mode) + str(self.__in_accesser_mode) +
            "True" + str(self.__allow_control) +
            '<SSH public key>')

        # Send information they need
        try:
            self.__net1_top.write(str(LOWEST_PROTOCOL_VERSION) + ',' + str(HIGHEST_PROTOCOL_VERSION) +
                                  str(self.__in_accesser_mode) + str(self.__in_accesser_mode) +
                                  'True' + str(self.__allow_control) +
                                  open(misc.CONFIG_DIR + "/local_rsa_key.pub", 'r').readline()) # End with a blank line
        except socket.error:
            self.__respond(callback_id, (1, _("Connection failed, please go back and try again")))
            self.__net1_top = None
            return

        start_time = time.time()
        self.__respond(callback_id, (0, _("Connected, swapping information...")), False)

        # Get the information they've sent
        max_wait_time = 10

        try:
            socket_string += self.__net1_top.read()
        except socket.error:
            pass

        while -1 == socket_string.find("\n"):

            if not self.__current_job_still_active(1):
                self.__net1_top = None
                return

            current_time = time.time() - start_time
            if current_time < max_wait_time:
                self.__respond(callback_id, (min(1, current_time / max_wait_time), None), False)
            else:
                self.__respond(callback_id, (1, _("Failed: the other computer connected but said nothing")))
                self.__net1_top = None
                return

            # Grab any waiting data
            try:
                socket_string += self.__net1_top.read()
            except socket.error:
                pass # Nothing to read

        debug.log("Received the following information:", socket_string)

        # Decompose the data they've sent
        socket_data = re.compile("([0-9]+),([0-9]+)(True|False)(True|False)(True|False)(True|False)ssh-rsa ([^ ]*) ((.*)@.*)\n").match(socket_string)

        if not socket_data:
            socket_data = re.compile("([0-9]+),([0-9]+)").match(socket_string)
            if socket_data:
                their_lowest_version, their_highest_version = socket_data.groups()
            else:
                self.__respond(callback_id, (1, _("Failed: couldn't understand the message sent by the other computer")))
                self.__net1_top = None
                return
        else:
            their_lowest_version, their_highest_version, they_are_in_accesser_mode, they_have_sshd, they_have_vnc, they_allow_control, \
                self.__their_public_key, self.__their_account, self.__their_username = socket_data.groups()

            if they_have_vnc:
                their_supported_protocols = [ "RFB" ]
            else:
                their_supported_protocols = [ "" ]

        debug.log(
            "- Decoded information as:",
            their_lowest_version + ',' + their_highest_version,
            they_are_in_accesser_mode, they_have_sshd, they_have_vnc, they_allow_control,
            '<SSH public key>')

        if int(their_lowest_version) > HIGHEST_PROTOCOL_VERSION:
            self.__respond(callback_id, (1, _("Failed: Their assistant is too new"),
                                       _("""The version of the remote help assistant they are using speaks a different language than the assistant you are using.

Please download a newer version of the assistant and try again.""")))
            self.__net1_top = None
            return
        elif int(their_highest_version) < LOWEST_PROTOCOL_VERSION:
            self.__respond(callback_id, (1, _("Failed: their assistant is too old"),
                                       _("""The version of the remote help assistant they are using speaks a different language than the assistant you are using.

Please ask them to download a newer version of the assistant and try again.""")))
            self.__net1_top = None
            return

        if "1" == their_highest_version:
            self.__net1_buffer = socket_string.split("\n", 1)[1]

            self.__connect_time = time.time()

        else: # Send/receive data not allowed in version 1

            self.__net1_top.write("RFB\n\n")
            
            while not socket_string.endswith("\n\n"):

                if not self.__current_job_still_active(1):
                    self.__net1_top = None
                    return

                current_time = time.time() - start_time
                if current_time < max_wait_time:
                    self.__respond(callback_id, (min(1, current_time / max_wait_time), None), False)
                else:
                    self.__respond(callback_id, (1, _("Failed: the other computer connected but said nothing")))
                    self.__net1_top = None
                    return

                # Grab any waiting data
                try:
                    socket_string += self.__net1_top.read()
                except socket.error:
                    pass # Nothing to read

            their_v2_protocols, self.__net1_buffer = socket_string.split("\n\n", 1)
            their_v2_protocols = their_v2_protocols.split("\n", 1)[1].split()
            if ( "RFB" in their_v2_protocols ) == ( "RFB" in their_supported_protocols ):
                their_supported_protocols = their_v2_protocols
            else:
                self.__respond(callback_id, (1, _("Failed: bug in the other side's assistant"),
                                                 _("""The assistant could not share access,
because of a bug in their assistant.

Please report this bug.""")))
                self.__net1_top = None
                return

                    

        self.__net1_top.blocking_reads()

        if ('True' == they_are_in_accesser_mode) and self.__in_accesser_mode:
            self.__respond(callback_id, (1, _("Failed: both sides tried to connect to each other"),
                                       _("""Both you and the other person tried to connect to each other's computers.
One of you needs to share their desktop instead.

Please go back and try again.""")))
            self.__net1_top = None
            return
        elif ('False' == they_are_in_accesser_mode) and not self.__in_accesser_mode:
            self.__respond(callback_id, (1, _("Failed: both sides tried to share access"),
                                       _("""Both you and the other person tried to share access to your computer.
One of you must connect to the other.

Please go back and try again.""")))
            self.__net1_top = None
            return

        elif ( 'False' == they_have_sshd ) and ( 'True' == they_are_in_accesser_mode ):
            self.__respond(callback_id, (1, _("Failed: SSH server needed"),
                                       _("""The assistant could not share access, because the other computer doesn't have an SSH server installed.  Please ask the other person to install an SSH server, then try again.""")))
            self.__net1_top = None
            return

        elif 'RFB' not in their_supported_protocols:
            self.__respond(callback_id, (1, _("Failed: no shared protocol"),
                                       _("""The version of the assistant they are using does not share any common desktop-sharing technologies""")))
            self.__net1_top = None
            return

        # Construct the success message from the data they've sent
        response = [ 1, _("Success!") ]

        ssh_keygen = subprocess.Popen(['/usr/bin/ssh-keygen', '-l', '-f', misc.CONFIG_DIR+'/local_rsa_key'], stdout=subprocess.PIPE)
        fingerprint_length, fingerprint = ssh_keygen.stdout.readline().rstrip().split(' ', 2)[0:2]
        ssh_keygen.wait()

        response.append(friendly_fingerprint(fingerprint))

        their_public_key_file = open(self.__current_key, 'w')
        print >> their_public_key_file, 'ssh-rsa', self.__their_public_key, self.__their_account
        their_public_key_file.close()
        ssh_keygen = subprocess.Popen(['/usr/bin/ssh-keygen', '-l', '-f', self.__current_key], stdout=subprocess.PIPE)
        fingerprint_length, fingerprint = ssh_keygen.stdout.readline().split(' ', 2)[0:2]
        ssh_keygen.wait()
        response.append(friendly_fingerprint(fingerprint))

        key_status = None

        account = self.__their_account

        if int(fingerprint_length) < 2048:
            key_status = 0
        else:
            if os.path.exists(misc.CONFIG_DIR + '/known_keys'):
                known_keys = open(misc.CONFIG_DIR + '/known_keys')

                for line in known_keys:
                    line = re.compile(" *([^ ]+) (.*?)( [0-7])?\n").match(line)
                    if line:
                        key, account, permissions = line.groups()
                        if account == self.__their_account:
                            self.__had_account = True
                            key_status = key == self.__their_public_key
                            if permissions is not None:
                                self.__previous_permissions = int(permissions)
                            break

        response.append(account)
        response.append(self.__previous_permissions)
        response.append(self.__required_permissions)

        if key_status:
            if self.__required_permissions == ( self.__required_permissions & self.__previous_permissions):
                response.append(_("""You have shared a session with this person before.
The next page will show their random dates,
and their level of trust.

You can safely skip the remaining pages"""))
                response.append(KEY_IS_GOOD)
            else:
                response.append(_("""You have not shared this type of session with this person before.
The next page will show you their random dates,
and options for their level of trust.

Please confirm their level of trust."""))
                response.append(KEY_NEEDS_PERMISSION)

        elif key_status is None:
            response.append(_("""You have never shared a session with this person before.
The next page will show you a series of random dates,
and options for this persion's level of trust.

Please telephone the other person,
and ask them to read out their dates.

When you have confirmed their dates,
please specify their level of trust.

Remember: criminals can fake e-mails and instant messages,
but can't impersonate people's voices."""))
            response.append(KEY_UNKNOWN)

        elif key_status is 0:
            response.append(_("""The identity they have sent is of the wrong type.
You are currently talking to an impostor - a criminal pretending to be someone you know.

Do not continue - please close the assistant."""))
            response.append(KEY_INVALID)

        else:
            response.append(_("""<b>WARNING:</b> Their identity has changed!
You may be talking to an impostor,
or they might just have changed their identity.

Please telephone the other person,
and ask them to read out their dates.

When you have confirmed their dates,
please specify their level of trust.

Criminals can fake e-mails and instant messages,
but can't impersonate people's voices."""))
            response.append(KEY_CHANGED)

        self.__respond(callback_id, response)


    ## Start a secure session with the remote assistant
    #
    # This function runs the second stage in the session: securing the
    # connection and starting a sharing server.
    #
    # The argument to this \ref backend_backend_commands "command" is
    # the set of permissions the account should be granted in future.
    # The callback will be called exactly once, with \c data
    # containing one of:
    #
    # \code
    # None                  # Failed to start the session
    # (None, error_message) # Failed with a reason
    # (server_type, port)   # In helper mode
    # server_type           # In sharer mode
    # \endcode
    #
    # At present, server_type is always 'RFB'.  'SHL' was supported in
    # earlier versions of the assistant, but was dropped due to lack
    # of use.  Other connection types might be added in future.
    #
    # In a VNC session, the helper is expected to start a VNC client
    # and connect to \c localhost:port.
    #
    # \section backend_backend_start_session_implementation Implementation details
    #
    # One end of the connection creates an SSH client, the other
    # an SSH server.  Client and server communicate over the
    # already-existing connection, even if it would be possible
    # for the SSH daemon to replace this program as the listener
    # on misc.REMOTE_PORT.
    #
    # This is for two reasons:
    #
    # - it avoids problems with misc.REMOTE_PORT being unavailable due
    #   to being in a FIN_WAIT2 or CLOSE_WAIT state
    #
    # - It simplifies the writing of the program (and therefore
    #   reduces the probability of bugs)
    #
    # The extra lag caused by bouncing the data repeatedly
    # between programs should be insignificant compared to
    # network lag.
    @misc.log_slow_functions
    def start_session(self, argument, callback_id):

        if self.__pid:
            return self.__send_backend_command('start_session', argument, callback_id)

        if self.__previous_permissions != argument:
            new_key = self.__their_public_key + ' ' + self.__their_account + ' ' + str(argument)
            self.__previous_permissions = argument

            if self.__had_account: # Remove all old instances of this key
                new_key += "\n"
                for line in open(misc.CONFIG_DIR + '/known_keys', 'r').readlines():
                    if line.split()[1] != self.__their_account:
                        new_key += line
                    open(misc.CONFIG_DIR + '/known_keys', 'w').write(new_key)
               
            else: # Just append this key
                print >> open(misc.CONFIG_DIR + '/known_keys', 'a'), new_key

        if self.__net1_top is None:
            return self.__respond(callback_id, (None, "Connect failed"))

        # Older clients don't notice the following write() if it's sent too quickly, so it has to be slowed down
        # A 1 second delay is too little, a 3 second delay should be enough
        if self.__connect_time is not None:
            sleep_time = time.time() - self.__connect_time
            if sleep_time < 3:
                time.sleep(3 - sleep_time)

        # Prove that we have verified the key
        # Note: a crash has been observed once on the next line, where self.__net1_top is None.
        # This might be fired by some odd combination of cancel_curent_job commands, but will be hard to repeat
        self.__net1_top.write("\n")

        if not self.__net1_buffer:
            while not select.select([ self.__net1_top ], [], [], 1)[0]:
                pass

        try:
            ssh_data = ( self.__net1_buffer + self.__net1_top.read() )[1:]
        except socket.error, msg:
            try:
                (dummy, errstr) = msg
                return self.__respond(callback_id, (None, errstr))
            except:
                return self.__respond(callback_id, (None, msg))

        if self.__in_accesser_mode:

            # We are running an SSH server, running this script in SSH
            # mode (see above).  Forward the remote SSH client to the
            # server and trap the I/O of the SSH mode script using
            # pipes.

            if os.path.exists(misc.CONFIG_DIR + "/from_ssh"): os.remove(misc.CONFIG_DIR + "/from_ssh")
            if os.path.exists(misc.CONFIG_DIR + "/to_ssh")  : os.remove(misc.CONFIG_DIR + "/to_ssh")
            if os.path.exists(misc.CONFIG_DIR + "/sshd.pid"):
                try:   os.kill(int(open(misc.CONFIG_DIR + "/sshd.pid", 'r').readline().rstrip()), 15)
                except OSError: pass
                try:   os.remove(misc.CONFIG_DIR + "/sshd.pid")
                except OSError: pass

            ## name of the sshd_config file
            config_file = misc.CONFIG_DIR + '/sshd_config'

            self.__join_pid = os.fork()
            if self.__join_pid:
                Backend.child_pids.append(self.__join_pid)
            else:
                debug.register_fork()
                os.dup2(self.__net1_top.fileno(), sys.stdout.fileno())
                os.dup2(self.__net1_top.fileno(), sys.stdin .fileno())

                ## Shell-escaped version of config_file
                config_dir_escaped = misc.CONFIG_DIR.replace('"', '\"')

                # Open, set permissions for, then reopen sshd_config, to avoid a race condition with the file being written to given an overly-permissive umask
                sshd_config = open(config_file, 'w')
                print >> sshd_config, 'temp'
                sshd_config.close()
                os.chmod(config_file, 0600)
                sshd_config = open(config_file, 'w')

                print >> sshd_config, \
"# This file was automatically created by", sys.argv[0], """, and will be recreated each time the program is run
HostKey \""""            + config_dir_escaped + """/local_rsa_key"
PidFile \""""            + config_dir_escaped + """/sshd.pid"
AuthorizedKeysFile \"""" + config_dir_escaped + """/current_key.""" + str(os.getppid()) + """"
AllowTcpForwarding no
ForceCommand exec""", misc.ASSISTANT_PATH + """/ssh_server_shell.py
IgnoreRhosts yes
PrintLastLog no
Port""", misc.REMOTE_PORT+1, """
ListenAddress 127.0.0.1
Protocol 2
LoginGraceTime 10
MaxStartups 1
PermitRootLogin no
PasswordAuthentication no
KeepAlive yes
ClientAliveInterval 60""" # """ # Silly emacs

                if debug.loglevel > 1:
                    print >> sshd_config, "LogLevel VERBOSE"
                sshd_config.close()

                os.execl("/usr/sbin/sshd", "/usr/sbin/sshd", '-i', '-D', '-f', config_file)

            while not os.path.exists(misc.CONFIG_DIR + "/from_ssh"):

                if os.waitpid(self.__join_pid, os.WNOHANG)[0]:
                    self.__respond(callback_id, None)
                    try: # Catch an error rather than checking whether os.path.exists() in case of a race condition
                        os.remove(self.__current_key)
                    except OSError:
                        pass
                    return

                if not self.__current_job_still_active(1):
                    os.kill(self.__join_pid, 15)
                    os.waitpid(self.__join_pid, 0)
                    try: # Catch an error rather than checking whether os.path.exists() in case of a race condition
                        os.remove(self.__current_key)
                    except OSError:
                        pass
                    return

            os.remove(config_file)

            self.__sec2_top = two_way_pipe.From_files_wr(misc.CONFIG_DIR + '/from_ssh', misc.CONFIG_DIR + '/to_ssh')

            ssh_data = ''
            while not ssh_data.endswith('\n'):
                try:
                    if select.select([ self.__sec2_top ], [], [], 1)[0]:
                        data = self.__sec2_top.read()
                        if not len(data):
                            self.__respond(callback_id, None)
                            return
                        ssh_data += data
                except select.error:
                    self.__respond(callback_id, None)
                    return

                if not self.__current_job_still_active():
                    return

            port = ssh_data.rstrip()

            # String or None (for unknown type)
            self.__respond(callback_id, ("RFB", port))

        else:

            # We are running an SSH client - trap its standard IO and
            # its network IO

            their_public_key_file = open(self.__current_key, 'w')
            their_public_key_file.write('* ssh-rsa ' + self.__their_public_key + "\n")
            their_public_key_file.flush()
            os.fsync(their_public_key_file.fileno())
            their_public_key_file.close()
            their_public_key_file = None

            global sigusr2_seen #pylint: disable-msg=I0011,W0603
            sigusr2_seen = False
            signal.signal(signal.SIGUSR2, handle_sigusr2)

            pid = os.fork()
            if pid:
                Backend.child_pids.append(pid)

                while not sigusr2_seen:
                    debug.log("Waiting for join_net1_sec2.py to begin...", os.getpid())
                    time.sleep(1)
            else:
                debug.register_fork()
                os.dup2(self.__net1_top.fileno(), sys.stdout.fileno())
                os.dup2(self.__net1_top.fileno(), sys.stdin .fileno())
                os.execl(misc.ASSISTANT_PATH + '/join_net1_sec2.py',
                         misc.ASSISTANT_PATH + '/join_net1_sec2.py', ssh_data
                         )

            self.__sec2_top = two_way_pipe.From_command(
                ( '/usr/bin/ssh',
                  #'-v', # debug the SSH connection
                  '-a', '-C', '-e', 'none', '-T', '-x',
                  '-i', misc.CONFIG_DIR + '/local_rsa_key',
                  '-p', str(misc.REMOTE_PORT+1), '-l', self.__their_username, 'localhost',
                  '-o', 'UserKnownHostsFile="' + self.__current_key.replace('"', '\"') + '"'
                  )
                )

            self.__app3_bottom = vnc_interaction.get_server(self.__allow_control)

            self.__respond(callback_id, "RFB")


    ## Run the session with the remote assistant
    # 
    # This function runs the final stage in the session: communication
    # over the secure connection.  This function cannibalises its
    # Backend instance, which can't process further commands after
    # this command begins.
    # 
    # The argument to this \ref backend_backend_commands "command" is
    # ignored.
    # 
    # The callback will be called exactly once, with \c data equal to
    # \c False if the connection provably exited prematurely, or \c
    # False if the connection \e might have finished normally.  Either
    # way, this signals the end of the session
    #
    # @note the front-end must not send any commands after main_loop()
    # is called.
    def main_loop(self, argument=None, callback_id=None):
        if self.__pid:
            self.__send_backend_command('main_loop', argument, callback_id)
            self.__save_config()
            return

        if self.__in_accesser_mode:

            # There's an SSH daemon doing the actual work - tell it to go
            self.__sec2_top.write("go\n")
            self.__sec2_top.blocking_reads()
            sec2_pid = int(self.__sec2_top.readline().rstrip())
            Backend.child_pids.append(sec2_pid)

            try:
                iwtd = []
                while not iwtd:
                    iwtd = select.select([ self.__sec2_top ], [], [], 10)[0]
                    if os.waitpid(self.__join_pid, os.WNOHANG)[0]:
                        os.kill(sec2_pid, signal.SIGTERM)
                        self.__message_pipe.close()
                        self.__message_pipe = None
                        return self.__respond(callback_id, False)
            except select.error:
                self.__message_pipe.close()
                self.__message_pipe = None
                return self.__respond(callback_id, False)

            self.__sec2_top.readline()

            self.__sec2_top.nonblocking_reads()
            self.__sec2_top.join(self.__message_pipe, 'Main session loop', verbose=False) # Die on command

        else:

            # Pipe data around the SSH client

            self.__sec2_top.nonblocking_reads()
            self.__app3_bottom.nonblocking_reads()
            self.__app3_bottom.join(self.__sec2_top, 'Main loop (back-end mode)', self.__message_pipe, verbose=True, conn1_name="app3 layer", conn2_name="sec2 layer")

            try: # checking whether os.path.exists() would be a race condition, so just catch the error
                os.remove(self.__current_key)
            except OSError:
                pass

        self.__respond(callback_id, True)

        self.__message_pipe.close()
        self.__message_pipe = None

# Set up CONFIG_DIR
if __name__ == '__main__':
    ## Debugging and logging
    debug = debugger.Debug()

    Backend()
else:
    ## Debugging and logging
    debug = debugger.Debug()

    if not os.path.exists(misc.CONFIG_DIR):
        if \
                os.mkdir(misc.CONFIG_DIR) or \
                os.chmod(misc.CONFIG_DIR, 0700) or \
                subprocess.call([ 'ssh-keygen', '-N', '', '-b', '2048', '-t', 'rsa', '-f', misc.CONFIG_DIR + "/local_rsa_key" ]):
    
            print _("Fatal error: couldn't set up %s") % misc.CONFIG_DIR
            sys.exit(1)

    os.chmod(misc.CONFIG_DIR, 0700)

    if debugger.LOGLEVEL < 2:
        sys.stderr = debugger.set_err_out(misc.CONFIG_DIR + "/send_with_bug_report.txt", True)
        debug.log("Starting with config dir:", misc.CONFIG_DIR)
        debug.log('revision data:', misc.VERSION_INFO)
