#!/usr/bin/python

## @package configurator_response.py
# Respond to a configurator file that has been detected
# \author Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>
#
# This responds to a configurator file, allowing the user to accept or
# reject the configuration, and to send a reply file.

## @file
# Simplified GTK client (sharing only)

## This file is distributed under the MIT license
LICENSE = """\
Copyright (c) 2008, Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os, sys, gettext, misc, backend
import pygtk
pygtk.require('2.0')
import gtk
import gtk.glade

gettext.install('remote-help-assistant')
gtk.glade.bindtextdomain('remote-help-assistant', sys.prefix + '/share/locale/')
gtk.glade.textdomain('remote-help-assistant')

## Quit the program
def gtk_main_quit(window, event=None):
    gtk.main_quit()

## Handle the user clicking on the "Help" button
def on_helpbutton_clicked(widget):
    try:
        import gnomevfs
        gnomevfs.url_show("ghelp:" + misc.DOC_PATH + '/manual.xml#rha-configuring')
    except ImportError:
        run_cmd = None

        if os.access("/usr/bin/sensible-browser", os.X_OK):
            run_cmd = 'sensible-browser "%s"'

        try:
            import gconf
            gconf_client = gconf.client_get_default()
            if gconf_client.get_bool('/desktop/gnome/url-handlers/http/enabled'):
                run_cmd = gconf_client.get_string('/desktop/gnome/url-handlers/http/command')
        except:
            pass

        if run_cmd is not None:
            if not os.fork():
                os.setsid()
                os.execl('/bin/sh', '/bin/sh', '-c', 'exec ' + run_cmd % ("ghelp:" + misc.DOC_PATH + '/manual.xml#rha-configuring'))

## Apply a single .rha file
def run(rha):
    def on_applybutton_clicked(widget):
        widget.apply_rha = True
        gtk.main_quit()

    ## Glade interface object
    interface = gtk.glade.XML(misc.ASSISTANT_PATH + '/configurator_response.glade')

    interface.signal_autoconnect({
            'gtk_main_quit'         : gtk_main_quit,
            'on_helpbutton_clicked' : on_helpbutton_clicked,
            'on_applybutton_clicked': on_applybutton_clicked,
            })

    widget = interface.get_widget( 'preamble') ; widget.set_markup(widget.get_label() % rha['their-account'])
    widget = interface.get_widget('postamble') ; widget.set_markup(widget.get_label() % rha['filename'])

    interface.get_widget('applybutton').apply_rha = False

    actions = []

    if 'choice-control' in rha:
        if rha['choice-control']: actions.append(_("&#x2022; allow keyboard and mouse control by default"))
        else                    : actions.append(_("&#x2022; not allow keyboard and mouse control by default"))

    if 'choice-firewall' in rha:
        if rha['choice-firewall']: actions.append(_("&#x2022; manage the firewall by default"))
        else                     : actions.append(_("&#x2022; not manage the firewall by default"))

    if 'choice-mode' in rha:
        actions.append({
                'force-safe': _("&#x2022; always use <i>safe mode</i>"),
                'default-safe': _("&#x2022; use <i>safe mode</i> by default"),
                'default-advanced': _("&#x2022; use <i>advanced mode</i> by default"),
                'force-advanced': _("&#x2022; always use <i>advanced mode</i>"),
                }[rha['choice-mode']])

    if 'choice-ip-editable' in rha:
        if rha['choice-ip-editable']:
            actions.append(_("&#x2022; connect to <i>%s</i> by default") % rha['choice-ip-address'])
        else:
            actions.append(_("&#x2022; always connect to <i>%s</i>") % rha['choice-ip-address'])

    if 'new-key' in rha:
        if 'changed-key' in rha and rha['changed-key']:
            actions.append(_("&#x2022; <b>replace</b> the identification information for <i>%s</i>") % rha['their-account'])

        if 'new-permissions' in rha:
            if rha['new-permissions'] & backend.TRUST_SHARE:
                actions.append(_("&#x2022; Allow %s to <i>view your screen</i>") % rha['their-account'])
            if rha['new-permissions'] & backend.TRUST_CONTROL:
                actions.append(_("&#x2022; Allow %s to <i>control your keyboard and mouse</i>") % rha['their-account'])
            if rha['new-permissions'] & backend.TRUST_ACCESS:
                actions.append(_("&#x2022; Allow %s to <i>share their screen with you</i>") % rha['their-account'])


    interface.get_widget('actions').set_markup("\n".join(actions))

    interface.get_widget('main-window').show()
    gtk.main()
    interface.get_widget('main-window').hide()

    return interface.get_widget('applybutton').apply_rha

## Save a user@host.rha configurator file
def save():
    def on_saveasbutton_clicked(widget):
        ssh_key = open(misc.CONFIG_DIR + "/local_rsa_key.pub", 'r').readline().strip()[8:]

        filechooser = gtk.FileChooserDialog(
            title=_("Save As"),
            parent=interface.get_widget("main-window"),
            action=gtk.FILE_CHOOSER_ACTION_SAVE,
            buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_SAVE, gtk.RESPONSE_OK)
            )
        filechooser.set_current_name(ssh_key.split()[1:] + ".rha")
        filechooser.set_do_overwrite_confirmation(True)

        if filechooser.run() == gtk.RESPONSE_OK:

            print >> open(filechooser.get_filename(), 'wb'), """#!/usr/bin/remote-help-assistant --config 
#
# This is a configuration file for the Remote Help Assistant
# (https://launchpad.net/remote-help-assistant).
#
# Although this can be executed as a script, it is not recommended:
# encouraging people to run scripts is a bad habit.
#
# This file includes an SSH public key, which must never be tampered
# with.  You should destroy this file if it has been sent over a
# non-tamper-proof link.  In particular:
#
#         NEVER SEND CONFIGURATOR FILES OVER THE INTERNET

[configurator]
ssh-key-permissions = access
ssh-key = """ + ssh_key
        gtk.main_quit()

    ## Glade interface object
    interface = gtk.glade.XML(misc.ASSISTANT_PATH + '/configurator_response_save.glade')

    interface.signal_autoconnect({
            'gtk_main_quit'         : gtk_main_quit,
            'on_helpbutton_clicked' : on_helpbutton_clicked,
            'on_saveasbutton_clicked': on_saveasbutton_clicked,
            })

    interface.get_widget('main-window').show()
    gtk.main()
    interface.get_widget('main-window').hide()
