## @package debugger
# Debugging classes and functions
# \author Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

## @file
# Debugging classes and functions

## This file is distributed under the MIT license
LICENSE = """\
Copyright (c) 2008, Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os, sys, time, inspect

## Amount of logging
#
# - 0: none
# - 1: enough to have a chance of a meaningful bug report
# - 2: enough to be useful during development
LOGLEVEL = 2

## Amount of testing
#
# - 0: none
# - 1: enough to test normal issues
# - 2: as 1, but disable anything that could e.g. use up resources on a remote server
#
# debug level 2 is useful when the program is being run repeatedly, to
# avoid using resources on some external website.
TESTLEVEL = 2

## Stream to send messages to
ERR_OUT = sys.stderr

## Directory to store pid files
PID_DIRECTORY = None

## Each module should use its own Debug object
class Debug:

    ## Constructor
    def __init__(self, loglevel_modifier=0, testlevel_modifier=0):
        ## Amount of logging to produce
        self.loglevel = LOGLEVEL + loglevel_modifier
        ## Amount of testing to do
        self.testlevel = TESTLEVEL

    ## Destructor
    def __del__(self):
        if PID_DIRECTORY:
            try:
                os.remove("%s/%d" % (PID_DIRECTORY, os.getpid()))
            except OSError:
                pass

    ## Print a debugging message
    #
    # Prints a message to the debugging output if loglevel is sufficiently high
    # The keyword argument "trace" causes a stack trace to be printed
    def log(self, *args, **kwargs):
        if self.loglevel > 0:

            stack_level = 1

            if 'stack_jump' in kwargs:
                stack_level += kwargs['stack_jump']

            filename, line = inspect.stack()[stack_level][1:3]

            filename = filename.split('/')[-1]

            message = filename + ':' + str(line) + ":\t" + \
                " ".join(map(lambda x: str(x), args)) + \
                " (at " + time.strftime("%a, %d %b %Y %H:%M:%S +0000 ", time.gmtime()) + ")"

            print >> ERR_OUT, message

            if 'trace' in kwargs and kwargs['trace']:
                import traceback
                print >> ERR_OUT, inspect.stack()[0][1].split('/')[-1] + ":" + str(inspect.stack()[0][2]) + ": Stack trace: ",
                traceback.print_exc(file=ERR_OUT)

    ## Mark a specific position in the program
    def mark(self):
        return self.log("MARKING FILE POSITION", stack_jump=1)

    ## Mark the start-point of a range in the program
    def mark_start(self):
        return self.log(">>> MARKING START OF SECTION", stack_jump=1)

    ## Mark the end-point of a range in the program
    def mark_end(self):
        return self.log("<<< MARKING END OF SECTION", stack_jump=1)

    ## Mark that the process has just forked a child (called from the child)
    def register_fork(self):
        if PID_DIRECTORY:
            open("%s/%d" % (PID_DIRECTORY, os.getpid()), 'wb').write("Process ID")

    ## Does the current test level indicate that we should contact other sites?
    #
    # While debugging, it's polite to reduce contact with computers
    # not associated with the session.
    def contact_external_sites(self):
        return self.testlevel < 2

## Set the file that error messages will be sent to from now on
def set_err_out(filename, truncate):
    global ERR_OUT #pylint: disable-msg=I0011,W0603
    if truncate:
        ERR_OUT = os.open(filename, os.O_CREAT |os.O_WRONLY)
        os.ftruncate(ERR_OUT, 0)
        os.chmod(filename, 0600)
        ERR_OUT = os.fdopen(ERR_OUT, 'wb', 0)
    else:
        ERR_OUT = os.open(filename, os.O_APPEND|os.O_WRONLY)
        ERR_OUT = os.fdopen(ERR_OUT, 'ab', 0)

    return ERR_OUT

## Set the directory where pid files will be saved to
def set_pid_directory(dirname):
    global PID_DIRECTORY #pylint: disable-msg=I0011,W0603
    PID_DIRECTORY = dirname

## Kill any PIDs with a file in the PID directory
def kill_pids():
    for pid in os.listdir(PID_DIRECTORY):
        if pid != os.getpid():
            try:
                os.kill(int(pid), 15)
            except OSError:
                pass
            try:
                os.remove(PID_DIRECTORY + '/' + pid)
            except OSError:
                pass

if 'RHA_DEBUGGER_FILE' in os.environ:
    set_err_out(os.environ['RHA_DEBUGGER_FILE'], False)

if 'RHA_DEBUGGER_LOGLEVEL' in os.environ:
    LOGLEVEL = int(os.environ['RHA_DEBUGGER_LOGLEVEL'])

if 'RHA_DEBUGGER_DIR' in os.environ:
    set_pid_directory(os.path.expanduser(os.environ['RHA_DEBUGGER_DIR']))
    open("%s/%d" % (PID_DIRECTORY, os.getpid()), 'wb').write("Process ID")
