## @package Gtk_advanced.py
# GTK interface class
# \author Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>
#
# This is the Remote Help Assistant's advanced GTK interface

## @file
# GTK advanced interface class

## This file is distributed under the MIT license
LICENSE = """\
Copyright (c) 2008, Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os, sys, backend, gettext, misc, subprocess, time
import pygtk
pygtk.require('2.0')
import gtk
import gtk.glade
import gobject, gconf
import debugger

gettext.install('remote-help-assistant')

## Debugging and logging
debug = debugger.Debug()

###############################
# CLIENT AND SERVER FUNCTIONS #
###############################

@misc.log_slow_functions
def run_browser(self, url, cmd):
    """Run a command with system(), then return immediately"""
    if not os.fork():
        os.setsid()
        os.execl('/bin/sh', '/bin/sh', '-c', 'exec ' + cmd % url.replace("ghelp:", "ghelp:" + misc.DOC_PATH + '/manual.xml'))

def glade_callback(function):
    """
    This decorator marks a function as a callback to a Glade signal
    Taken from http://blog.rabidgeek.com/?p=113
    """
    function.__glade_callback__ = True
    return function

class Gtk_remote_help_assistant(object):
    """GTK assistant

    This is the main class for the program's GTK mode, definining
    the assistant and its behaviour
    """

    ## Constructor
    def __init__(self, back_end):

        gtk.glade.bindtextdomain('remote-help-assistant', sys.prefix + '/../locale/')
        gtk.glade.textdomain('remote-help-assistant')

        ## Glade interface object
        self.interface = gtk.glade.XML(misc.ASSISTANT_PATH + '/gtk_advanced.glade')

        ## Main GTK window
        self.window = self.interface.get_widget("main-window")

        glade_callbacks = {}
        for key in dir(self):
            attr = getattr(self, key)
            if callable(attr) and hasattr(attr, '__glade_callback__'):
                glade_callbacks[key] = attr

        self.interface.signal_autoconnect(glade_callbacks)

        self.window.add_events(gtk.gdk.ENTER_NOTIFY_MASK | gtk.gdk.CONFIGURE)

        ## Back-end instance
        self.backend = back_end
        ## Timer for polling the message queue
        self.message_queue_handler = gobject.timeout_add(1000, self.poll_message_queue)

        gconf_client = gconf.client_get_default()

        try:
            import gnomevfs
            gtk.link_button_set_uri_hook(lambda self, url: gnomevfs.url_show(url.replace("ghelp:", "ghelp:" + misc.DOC_PATH + '/manual.xml')))
        except ImportError:
            if gconf_client.get_bool('/desktop/gnome/url-handlers/http/enabled'):
                ## Web browser command
                self.browser = gconf_client.get_string('/desktop/gnome/url-handlers/http/command')
                gtk.link_button_set_uri_hook(run_browser, self.browser)
            elif os.access("/usr/bin/sensible-browser", os.X_OK):
                self.browser = 'sensible-browser "%s"'
                gtk.link_button_set_uri_hook(run_browser, self.browser)

        run_vnc_client = self.interface.get_widget('vnc-autostart')
        if gconf_client.get_bool('/desktop/gnome/url-handlers/vnc/enabled'):
            ## VNC client command
            self.vnc_client = gconf_client.get_string('/desktop/gnome/url-handlers/vnc/command')
            run_vnc_client.set_active(True)
        else:
            self.vnc_client = ''
            run_vnc_client.set_active(False)
            run_vnc_client.set_sensitive(False)


        ## the assistant is said to be "in helper mode" if it's accessing another computer, or "sharer mode" otherwise
        self.__in_helper_mode = False

        ## Is there an insecure connection available that hasn't yet had any data written to it?
        self.connection_established = False

        ## Is there a session currently active?
        self.active_session = False

        ## Timer that pulses the current progress bar (if any)
        self.pulser = None

        ## Permissions associated with the current key
        self.__key_permissions = None
        ## Permissions required in order to continue
        self.__required_permissions = None

        ## Port to point the VNC client at (if any)
        self.vnc_port = False

        ## Set the IP address description loader going
        self.backend.get_network_access(None, self.set_network_access)

        ## Main "their IP address" entry box
        self.remote_IP_simple = self.interface.get_widget('page3-their-address-simple-entry')

        addressbox_simple = self.interface.get_widget('page3-their-address-simple')
        addressbox_simple.set_model(gtk.ListStore(gobject.TYPE_STRING))
        addressbox_simple.set_text_column(0)
        addressbox_simple.pack_start(gtk.CellRendererText(), True)

        # Fill the dropdown box
        for address in self.backend.get_recently_shared():
            addressbox_simple  .append_text(address)

        addressbox_simple.set_active(0)
        addressbox_simple.set_sensitive(self.backend.get_default_ip_editable())

        ## Advanced "their IP address" entry box
        self.remote_IP_advanced = self.interface.get_widget('page3-their-address-advanced-entry')

        addressbox_advanced = self.interface.get_widget('page3-their-address-advanced')
        addressbox_advanced.set_model(gtk.ListStore(gobject.TYPE_STRING))
        addressbox_advanced.set_text_column(0)
        addressbox_advanced.pack_start(gtk.CellRendererText(), True)

        # Fill the dropdown box
        for address in self.backend.get_recently_accessed():
            addressbox_advanced.append_text(address)

        addressbox_advanced.set_sensitive(self.backend.get_default_ip_editable())

        ## Widget used at various points throughout the program
        self.iptables          = self.interface.get_widget('page3-manage-firewall')
        ## Widget used at various points throughout the program
        self.progress_bar      = self.interface.get_widget('page4-progress')

        ## Widgets that should be desensitised while a connection is in progress
        self.noprogress_items = (
            addressbox_simple,
            addressbox_advanced,
            self.iptables
            )

        ## Currently-visible final page label
        self.final_page_label = self.interface.get_widget('final-page-wait')

        ## Set pages complete that don't need special treatment
        self.window.set_page_complete(self.window.get_nth_page(0), True)
        self.window.set_page_complete(self.window.get_nth_page(1), True)

        if not back_end.get_interface_change_allowed():
            self.interface.get_widget('final-page-safe-mode').hide()

        if self.backend.get_access_mode():
            self.interface.get_widget('mode-access'       ).set_active(True)
        elif self.backend.get_default_control():
            self.interface.get_widget('mode-share-control').set_active(True)
        else:
            self.interface.get_widget('mode-share'        ).set_active(True)

        self.iptables.set_active(self.backend.get_default_firewall())

        ## Did the program exit after the "Apply" button was pressed?
        self.apply = False

    #@misc.log_slow_functions # This is bound to be slow, but the GUI has already closed so it doesn't matter
    ## Handle the main loop
    def after_close(self):

        if not self.apply:
            return

        if self.vnc_port:
            # Detach this from the rest of the assistant, so that it's not killed with the rest of the program
            if self.vnc_client and not os.fork():
                os.setsid()
                subprocess.Popen('exec ' + self.vnc_client % '"localhost:' + self.vnc_port + '"', shell=True)
                sys.exit(0)

            while not self.backend.check_set_firewall() and not self.backend.has_died():
                time.sleep(1)

        elif self.vnc_port is None:
            def handle_dialog(self, dialog):
                if self.backend.check_set_firewall() or self.backend.has_died():
                    dialog.response(0)
                return True

            secondary_text = _("""You are currently sharing your desktop with somebody else. They will have access to your desktop until this window is closed.

If you are unable to close this window for some reason, you can force your session to quit by holding down <i>ctrl-alt-backspace</i>. The <i>backspace</i> key is labelled <i>&#x2190;</i> on most keyboards, and is the wide key above the <i>return</i> key (<i>&#x23ce;</i>).""")
            if self.iptables.get_active():
                secondary_text += _("""

<i>Changes made automatically to your firewall will be undone when this window closes. You might be asked to enter your password.</i>""")

            dialog = self.dialog(_("Desktop being shared"), _("To stop sharing your desktop,\nclose this window"), secondary_text)
            handler = gobject.timeout_add(1000, handle_dialog, self, dialog)
            ret = dialog.run() # returns 0 if the backend died
            dialog.hide()
            gobject.source_remove(handler)
            return ret

        elif 0 == self.vnc_port:
            return

        if self.iptables.get_active():
            dialog = self.dialog(_("Remote connection closed"), _('Remote connection closed'),
                                 _("""The remote help session has now been closed.

<i>Changes made automatically to your firewall will now be undone. You might be asked to enter your password.</i>"""))
            dialog.run()
            dialog.hide()


    @misc.log_slow_functions
    def poll_message_queue(self):
        """Read any waiting callbacks, and complain if the backend dies"""
        if not self.backend.poll_message_queue():
            print >> sys.stderr, misc.now(), '- Got error from poll_message_queue(), exiting'

            dialog = gtk.MessageDialog(parent=self.window, flags=gtk.DIALOG_MODAL, type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_CLOSE, message_format=
_("""The assistant has crashed because of a bug.

The remote help assistant will now close.
This was caused by a design flaw in the assistant, and is not your fault.

Please report this error by filing a bug report through the website you downloaded the assistant from.

Please include the following file in your bug report:

%s/send_with_bug_report.txt""") % misc.CONFIG_DIR)
            dialog.set_title(_("The assistant has crashed"))
            dialog.run()
            gtk.main_quit()
            raise self.backend # To make apport run, if it's listening

        return True

    @glade_callback
    @misc.log_slow_functions
    def on_cancel(self, assistant, event=None):
        """Clean up the assistant when it's cancelled."""
        gobject.source_remove(self.message_queue_handler)
        self.backend.cancel_current()
        self.window.hide()
        gtk.main_quit()

    @glade_callback
    @misc.log_slow_functions
    def on_close(self, assistant):
        """When the user clicks 'apply', prepare for the main loop"""

        gobject.source_remove(self.message_queue_handler)
        self.window.hide()
        gtk.main_quit()
        self.backend.main_loop()

    @glade_callback
    @misc.log_slow_functions
    def on_enter_notify(self, widget, event=None):
        """Disable the urgency hint when the window is focussed"""
        self.window.set_urgency_hint(False)

    ## Set the text displayed on the final page
    def set_final_page_label(self, name, data=None):
        final_page_label = self.interface.get_widget(name)
        if self.final_page_label != final_page_label:
            self.final_page_label.hide()
            self.final_page_label = final_page_label
            if data is not None:
                if not hasattr(final_page_label, '__original_text__'):
                    final_page_label.__original_text__ = final_page_label.get_label()
                final_page_label.set_markup(final_page_label.__original_text__ % data)
            self.final_page_label.show()
        return final_page_label


    @misc.log_slow_functions
    def dialog(self, title, primary_text, secondary_text):
        """Creates a not-quite-dialog box

        The assistant sometimes needs to pop up a not-quite-dialog
        box.  This is a dialog box with a title, that shows up in
        the pager and taskbar.  This is used to convey information
        once the main window has closed."""

        dialog = gtk.MessageDialog(type=gtk.MESSAGE_INFO, buttons=gtk.BUTTONS_CLOSE, message_format=primary_text)
        dialog.format_secondary_markup(secondary_text)

        dialog.set_title(title)
        dialog.set_property("skip-taskbar-hint", False)
        dialog.set_property("skip-pager-hint"  , False)

        return dialog

    @glade_callback
    @misc.log_slow_functions
    def on_prepare(self, assistant, page):
        """Prepare the next page for display"""

        page_num = assistant.get_current_page()

        debug.log("Preparing page", page_num, '-', assistant.get_page_title(page))

        if page_num > 1:
            self.backend.cancel_current()

        if self.pulser is not None:
            gobject.source_remove(self.pulser)
            self.pulser = None

        if 2 == page_num:

            # General information
            if self.__in_helper_mode:
                whose_widgets = (
                    'page3-whose-computer-simple'  , 'page3-your-address-simple' , 'page3-their-address-advanced',
                    'page3-whose-computer-advanced', 'page3-their-address-simple', 'page3-your-address-advanced'
                    )

                self.interface.get_widget('page3-noconnect').hide()
                self.window.set_page_complete(page, True)
            else:
                whose_widgets = (
                    'page3-whose-computer-advanced', 'page3-their-address-simple', 'page3-your-address-advanced',
                    'page3-whose-computer-simple'  , 'page3-your-address-simple' , 'page3-their-address-advanced'
                    )

                widget = self.interface.get_widget('page3-noconnect')
                widget.show()
                self.remote_IP_simple.grab_focus()

                self.window.set_page_complete(self.window.get_nth_page(2), len(self.remote_IP_simple.get_text()) or widget.get_active())

            widget = self.interface.get_widget(whose_widgets[0])
            widget.set_markup(_("<b>Internet address of this computer</b>"))
            self.interface.get_widget(whose_widgets[1]).show()
            self.interface.get_widget(whose_widgets[2]).show()

            widget = self.interface.get_widget(whose_widgets[3])
            widget.set_markup(_("<b>Internet address of the other computer</b>"))
            self.interface.get_widget(whose_widgets[4]).hide()
            self.interface.get_widget(whose_widgets[5]).hide()

        elif 3 == page_num and not self.connection_established:
            self.window.set_page_complete(page, False)

            self.interface.get_widget('page4-warning').hide()
            self.progress_bar.show()
            self.progress_bar.set_fraction(0)
            self.progress_bar.set_text(_("Please wait..."))
            self.progress_bar.set_pulse_step(0.01)
            self.pulser = gobject.timeout_add(50, self.pulse, self.progress_bar)

            allow_control = self.interface.get_widget('mode-share-control').get_active()

            if self.__in_helper_mode:
                if self.interface.get_widget('page3-noconnect').get_active():
                    address = ""
                else:
                    address = self.remote_IP_advanced.get_text()
                self.backend.connect((True , address                         , allow_control, self.iptables.get_active()), self.handle_connection)
            else:
                self.backend.connect((False, self.remote_IP_simple.get_text(), allow_control, self.iptables.get_active()), self.handle_connection)

        elif 6 == page_num and not self.active_session:
            # Confirm session creation
            self.set_final_page_label('final-page-wait')

            self.backend.start_session(self.__key_permissions, self.handle_session)
            self.active_session = True

        debug.log("Done preparing page.")

    ## Handle the user changing the address to connect to
    @glade_callback
    @misc.log_slow_functions
    def on_address_changed(self, widget):
        if not self.__in_helper_mode:
            self.window.set_page_complete(self.window.get_nth_page(2), bool(self.remote_IP_simple.get_text()))

    ## Handle the user changing the "do not connect" button
    @glade_callback
    @misc.log_slow_functions
    def on_noconnect_toggled(self, widget):
        if not self.__in_helper_mode:
            self.interface.get_widget('page3-their-address-simple').set_sensitive(not widget.get_active())
            self.window.set_page_complete(self.window.get_nth_page(2), bool(self.remote_IP_simple.get_text()) or widget.get_active())

    @glade_callback
    @misc.log_slow_functions
    def on_helper_mode_toggled(self, widget):
        """Switch between accessing a computer and sharing access to this one"""
        self.__in_helper_mode = widget.get_active()

    ## Handle the user changing the cybercrime button
    @glade_callback
    @misc.log_slow_functions
    def on_page5_crime_toggled(self, widget):
        self.window.set_page_complete(self.window.get_nth_page(4), widget.get_active())

    @misc.log_slow_functions
    def set_network_access(self, data):
        """Set the message on page 3 describing this computer's IP address"""

        success, message, pronunciation, router_URL, private_address = data

        label = self.interface.get_widget('page4-warning')

        if success:
            label.hide()
            self.interface.get_widget('page3-your-address-simple').set_markup(
                _("Your Internet address is %s\n(<i>%s</i>)")
                % ( message, pronunciation )
                )
            self.interface.get_widget('page3-your-address-advanced').set_markup(
                _("Your Internet address is %s\n(<i>%s</i>)")
                % ( message, pronunciation )
                )
            self.interface.get_widget('page3-private-address').set_markup(
                _("Your private address is %s\nThe assistant connects to TCP port number %d")
                % (private_address, misc.REMOTE_PORT))
        else:
            label.set_text(message)
            label.show()

        button = self.interface.get_widget('page3-router-config')
        button.set_sensitive("" != router_URL)
        button.set_uri(router_URL)
        button.set_tooltip_text(router_URL)

    @misc.log_slow_functions
    def pulse(self, pbar):
        """Pulse the progress bar"""
        pbar.pulse()
        return True

    @misc.log_slow_functions
    def handle_connection(self, data):
        """Callback for the connect() command """

        if data[0] is not None:
            if self.pulser is not None:
                gobject.source_remove(self.pulser)
                self.pulser = None

            self.progress_bar.set_fraction(data[0])

        if data[1] is not None:
            self.progress_bar.set_text(data[1])

        if 2 == len(data):
            return

        if 3 == len(data):
            warning = self.interface.get_widget('page4-warning')
            warning.show()
            warning.set_markup(data[2])

        elif 5 == len(data):
            self.set_network_access(data)

        elif 9 != len(data):
            print >> sys.stderr, misc.now(), "- Impossible data sent to handle_connection():", data
            raise data

        else:

            our_key, their_key, account, self.__key_permissions, self.__required_permissions, their_message, key_validity = data[2:]

            widget = self.interface.get_widget('page5-crime')
            widget.set_active(backend.KEY_CHANGED != key_validity and
                              backend.KEY_INVALID != key_validity)
            self.window.set_page_complete(self.window.get_nth_page(4),
                                          backend.KEY_CHANGED != key_validity and
                                          backend.KEY_INVALID != key_validity)

            image = self.interface.get_widget('page5-image')

            if   backend.KEY_IS_GOOD == key_validity: widget.hide() ; image.set_from_file(misc.ASSISTANT_PATH + '/identity-good.svg')
            elif backend.KEY_UNKNOWN == key_validity: widget.hide() ; image.set_from_file(misc.ASSISTANT_PATH + '/identity-new.svg')
            else                                    : widget.show() ; image.set_from_file(misc.ASSISTANT_PATH + '/identity-bad.svg')

            self.interface.get_widget('page5-description').set_label(their_message)
            self.interface.get_widget('page6-yourdates').set_label(our_key)
            self.interface.get_widget('page6-theirdates').set_label(their_key)

            widget = self.interface.get_widget('page6-trust-heading')
            if not hasattr(widget, '__original_text__'):
                widget.__original_text__ = widget.get_label()
                widget.set_markup(widget.__original_text__ % account)

            self.page6_set_trust(True, 0)

            self.connection_established = True

            tooltips = gtk.Tooltips()
            for widget in (
                self.interface.get_widget('mode-share'),
                self.interface.get_widget('mode-share-control'),
                self.interface.get_widget('mode-access'),
                ):
                widget.set_sensitive(False)
                tooltips.set_tip(widget, _("This can't be changed once the connection has been established"))

            self.window.set_page_complete(self.window.get_nth_page(3), True)
            if not self.window.is_active():
                self.window.set_urgency_hint(True)

    ## Set the permissions for the other person
    @misc.log_slow_functions
    def page6_set_trust(self, add_permissions, mask):
        if add_permissions:
            self.__key_permissions |= mask
        else:
            self.__key_permissions &= ~mask

        self.interface.get_widget('page6-trust-access'       ).set_active(self.__key_permissions & backend.TRUST_ACCESS)
        self.interface.get_widget('page6-trust-share'        ).set_active(self.__key_permissions & backend.TRUST_SHARE)
        self.interface.get_widget('page6-trust-share-control').set_sensitive(self.__key_permissions & backend.TRUST_SHARE)
        self.interface.get_widget('page6-trust-share-control').set_active(self.__key_permissions & backend.TRUST_CONTROL)
        self.window.set_page_complete(self.window.get_nth_page(5),
                                      ( self.__key_permissions & self.__required_permissions ) == self.__required_permissions
                                      )

    ## Handle the user changing the "trust-to-share" button
    @glade_callback
    @misc.log_slow_functions
    def page6_trust_share_clicked(self, widget):
        self.page6_set_trust(widget.get_active(), backend.TRUST_SHARE)

    ## Handle the user changing the "trust-to-control" button
    @glade_callback
    @misc.log_slow_functions
    def page6_trust_control_clicked(self, widget):
        self.page6_set_trust(widget.get_active(), backend.TRUST_CONTROL)

    ## Handle the user changing the "trust-to-access" button
    @glade_callback
    @misc.log_slow_functions
    def page6_trust_access_clicked(self, widget):
        self.page6_set_trust(widget.get_active(), backend.TRUST_ACCESS)

    ## Handle the user changing the "Apply" button
    @glade_callback
    @misc.log_slow_functions
    def on_apply(self, widget):
        if not self.backend.config.has_section('gtk'):
            self.backend.config.add_section('gtk')
        if self.interface.get_widget('final-page-safe-mode').get_active():
            self.backend.set_default_interface("safe")
        else:
            self.backend.set_default_interface("advanced")
        self.apply = True


    @misc.log_slow_functions
    def handle_session(self, data):
        """Callback for the start_session() command"""

        # Final page
        if data is None:
            self.set_final_page_label('final-page-error', misc.CONFIG_DIR)
            return

        if data[0] is None:
            print >> sys.stderr, misc.now(), "- Final page problem:", data[1]
            self.set_final_page_label('final-page-problem', data[1])
            return

        if self.__in_helper_mode:
            self.vnc_port = data[1]
            self.set_final_page_label('final-page-helper-vnc-container')
            label = self.interface.get_widget('final-page-helper-vnc')
            if not hasattr(label, '__original_text__'):
                label.__original_text__ = label.get_label()
            label.set_markup(label.__original_text__ % data[1])

            if not self.vnc_client:
                self.interface.get_widget('vnc-autostart').set_active(False)

        else:
            self.vnc_port = None
            self.set_final_page_label('final-page-helpee-vnc')

        self.window.set_page_complete(self.window.get_nth_page(6), True)

## Run the assistant
def run(back_end):
    assistant = Gtk_remote_help_assistant(back_end)
    gtk.main()
    assistant.after_close()
