#!/usr/bin/python

## @package gtk_safe.py
# Simplified GTK client (sharing only)
# \author Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>
#
# This is the Remote Help Assistant's simplified GTK interface

## @file
# Simplified GTK client (sharing only)

## This file is distributed under the MIT license
LICENSE = """\
Copyright (c) 2008, Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import sys, gettext, misc, backend, gtk_advanced
import pygtk
pygtk.require('2.0')
import gtk
import gtk.glade
import gobject

gettext.install('remote-help-assistant')

def glade_callback(function):
    """
    This decorator marks a function as a callback to a Glade signal
    Taken from http://blog.rabidgeek.com/?p=113
    """
    function.__glade_callback__ = True
    return function

## Safe interface to the assistant
class Gtk_safe(object):

    ## Constructor
    def __init__(self, back_end):

        gtk.glade.bindtextdomain('remote-help-assistant', sys.prefix + '/share/locale/')
        gtk.glade.textdomain('remote-help-assistant')

        ## Glade interface object
        self.interface = gtk.glade.XML(misc.ASSISTANT_PATH + '/gtk_safe.glade')

        ## Main GTK window
        self.window = self.interface.get_widget("main-window")

        glade_callbacks = {}
        for key in dir(self):
            attr = getattr(self, key)
            if callable(attr) and hasattr(attr, '__glade_callback__'):
                glade_callbacks[key] = attr

        self.interface.signal_autoconnect(glade_callbacks)

        ## Back-end instance
        self.backend = back_end
        ## Timer for polling the message queue
        self.message_queue_handler = gobject.timeout_add(1000, self.poll_message_queue)

        ## Timer that pulses the current progress bar (if any)
        self.pulser = None

        ## Main "their IP address" entry box
        self.remote_ip_address = self.interface.get_widget('their-address-entry')

        addressbox = self.interface.get_widget('their-address')
        addressbox.set_model(gtk.ListStore(gobject.TYPE_STRING))
        addressbox.set_text_column(0)
        addressbox.pack_start(gtk.CellRendererText(), True)

        # Fill the dropdown box
        for address in self.backend.get_recently_shared():
            addressbox  .append_text(address)

        addressbox.set_active(0)
        addressbox.set_sensitive(self.backend.get_default_ip_editable())

        ## Widget used at various points throughout the program
        self.allow_control     = self.interface.get_widget('control')
        ## Widget used at various points throughout the program
        self.iptables          = self.interface.get_widget('firewall')
        ## Widget used at various points throughout the program
        self.progress_bar      = self.interface.get_widget('progress')

        ## Should we run advanced mode after exiting?
        self.run_advanced_mode = False

        if not back_end.get_interface_change_allowed():
            self.interface.get_widget('advanced').hide()

        self.allow_control.set_active(self.backend.get_default_control())
        self.iptables     .set_active(self.backend.get_default_firewall())

        ## Widgets that should be desensitised while a connection is in progress
        self.noprogress_items = (
            addressbox,
            self.allow_control,
            self.iptables,
            )

        ## Widgets that should be denesnitised when a connection succeeds
        self.nocomplete_items = (
            self.interface.get_widget('connectcancel'),
            self.interface.get_widget('advanced'),
            )

        self.window.show()

    @misc.log_slow_functions
    def poll_message_queue(self):
        """Read any waiting callbacks, and complain if the backend dies"""
        if not self.backend.poll_message_queue():
            print >> sys.stderr, misc.now(), '- Got error from poll_message_queue(), exiting'

            dialog = gtk.MessageDialog(parent=self.window, flags=gtk.DIALOG_MODAL, type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_CLOSE, message_format=
_("""The assistant has crashed because of a bug.

The remote help assistant will now close.
This was caused by a design flaw in the assistant, and is not your fault.

Please report this error by filing a bug report through the website you downloaded the assistant from.

Please include the following file in your bug report:

%s/send_with_bug_report.txt""") % misc.CONFIG_DIR)
            dialog.set_title(_("The assistant has crashed"))
            dialog.run()
            gtk.main_quit()
            raise self.backend # To make apport run, if it's listening

        return True

    ## Handle the user clicking on the "Advanced" button
    @glade_callback
    @misc.log_slow_functions
    def on_advanced_clicked(self, window, event=None):
        self.backend.cancel_current()
        self.run_advanced_mode = True
        return self.on_close(window, event)

    ## Handle the user clicking on the "Close" button
    @glade_callback
    @misc.log_slow_functions
    def on_close(self, window, event=None):
        """Clean up the assistant when it's closed."""
        if self.message_queue_handler is not None:
            gobject.source_remove(self.message_queue_handler)
        self.backend.cancel_current()
        self.window.hide()
        gtk.main_quit()


    ## Handle the user clicking on the "Connect" button
    @glade_callback
    @misc.log_slow_functions
    def on_connectbutton_clicked(self, widget):
        """Begin connecting to the other assistant"""
        self.interface.get_widget('connectbutton').hide()
        self.interface.get_widget('connectbox-inner').show()
        self.interface.get_widget('connectcancel').grab_focus()
        self.interface.get_widget('warning').hide()
        self.progress_bar.set_fraction(0)
        self.progress_bar.set_text(_("Please wait..."))
        self.progress_bar.set_pulse_step(0.01)
        self.pulser = gobject.timeout_add(50, self.pulse, self.progress_bar)

        for widget in self.noprogress_items:
            widget.set_sensitive(False)

        self.backend.connect((False, self.remote_ip_address.get_text(), self.allow_control.get_active(), self.iptables.get_active()), self.handle_connection)

    ## Handle the user clicking on the "cancel" button
    @glade_callback
    @misc.log_slow_functions
    def on_connectcancel_clicked(self, widget):
        """Cancel connecting"""
        self.interface.get_widget('connectbox-inner').hide()
        self.interface.get_widget('connectbutton').show()
        self.interface.get_widget('connectbutton').grab_focus()
        self.backend.cancel_current()
        for widget in self.noprogress_items:
            widget.set_sensitive(True)
        if self.pulser is not None:
            gobject.source_remove(self.pulser)
            self.pulser = None

    @misc.log_slow_functions
    def pulse(self, pbar):
        """Pulse the progress bar"""
        pbar.pulse()
        return True

    @misc.log_slow_functions
    def handle_connection(self, data):
        """Callback for the connect() command """

        if data[0] is not None:
            if self.pulser is not None:
                gobject.source_remove(self.pulser)
                self.pulser = None

            self.progress_bar.set_fraction(data[0])

        if data[1] is not None:
            self.progress_bar.set_text(data[1])

        if 2 == len(data):
            return

        if 3 == len(data):
            warning = self.interface.get_widget('warning')
            warning.show()
            warning.set_markup(data[2])

        elif 5 == len(data):

            if not data[0]:
                warning = self.interface.get_widget('warning')
                warning.show()
                warning.set_markup(data[1])

        elif 9 != len(data):
            print >> sys.stderr, misc.now(), "- Impossible data sent to handle_connection():", data
            raise data

        else:

            if backend.KEY_IS_GOOD == data[8]:
                for widget in self.nocomplete_items:
                    widget.set_sensitive(False)

                self.backend.start_session(data[5], self.handle_session)

                if not self.window.is_active():
                    self.window.set_urgency_hint(True)

            else:
                self.progress_bar.set_text(_("Failed: identity not recognised"))
                warning = self.interface.get_widget('warning')
                warning.show()
                warning.set_markup(_("""The assistant couldn't confirm the identity of the other person.
Please try again, in advanced mode."""))

    @glade_callback
    @misc.log_slow_functions
    def on_enter_notify(self, widget, event=None):
        """Disable the urgency hint when the window is focussed"""
        self.window.set_urgency_hint(False)

    @misc.log_slow_functions
    def handle_session(self, data):
        """Callback for the start_session() command"""

        # Final page
        if data is None or data[0] is None:
            self.progress_bar.set_text(_("Failed: identity not recognised"))
            warning = self.interface.get_widget('warning')
            warning.show()
            warning.set_markup(_("""The assistant couldn't confirm the identity of the other person.
Please try again, in advanced mode."""))
        else:
            self.progress_bar.set_text(_("Now sharing your desktop"))
            self.interface.get_widget('close').grab_focus()
            self.backend.main_loop(None, self.handle_loop_end)

    @misc.log_slow_functions
    def handle_loop_end(self, data):
        """Callback for the main_loop() command"""

        self.progress_bar.set_text(_("Session finished."))
        gobject.source_remove(self.message_queue_handler)
        self.message_queue_handler = None
        


## Run the assistant
def run(back_end):
    assistant = Gtk_safe(back_end)
    gtk.main()

    if assistant.run_advanced_mode:
        assistant = None
        gtk_advanced.run(back_end)
