## @package two_way_pipe
# Classes for two-way communication between endpoints
# \author Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>
#
# A large part of the back-end's job is to funnel data between
# different streams.  This module presents a common interface to
# different types of network and file streams.

## @file
# two-way pipe package

## This file is distributed under the MIT license
LICENSE = """\
Copyright (c) 2008, Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import fcntl, os, select, socket, subprocess, sys, time
import debugger

## Debugging and logging
debug = debugger.Debug()

## Wraps a pair of pipes in a useful interface
#
# At several points in this program, it's necessary to have an object with the following properties:
# 
# - A non-blocking read()
# - A reliable, blocking write()
# - a fileno() that returns a readable fd
# 
# Non-blocking read()s and blocking write()s are needed so that the
# program can read and write whatever data's available without
# worrying about a deadlock or write failure.  A fileno() is needed
# for functions like select().
class From_filehandles:

    ## Constructor
    #
    # @param self this connection
    # @param from_fh filehandle to read from
    # @param to_fh filehandle to write to
    # @param blocking_reads should the pipe block while reading?
    def __init__(self, from_fh, to_fh, blocking_reads=False):
        ## Filehandle of the writing end of the pipe
        self.__output = to_fh
        ## Filehandle of the reading end of the pipe
        self.__input  = from_fh

        if not blocking_reads: self.nonblocking_reads()

    ## Filehandle used for reading
    def reader(self): return self.__input

    ## Filehandle used for writing
    def writer(self): return self.__output

    ## Close both ends of the pipe
    def close(self):
        self.__output.close()
        self. __input.close()

    ## Has the pipe been closed?
    def closed(self):
        return self.__output.closed or self.__input.closed

    ## File number for reading 
    def fileno(self):
        return self.__input.fileno()

    ## Write a string to the pipe
    def write(self, string):
        self.__output.write(string)
        self.__output.flush()

    ## Read up to 4096 bytes from the pipe
    #  @param self From_filehandles instance
    #  @param count maximum number of bytes to read
    #  @param blocking whether to block on reads
    #
    # @note: \c blocking is not supported on all types of socket
    def read(self, count=4096, blocking=0):
        return self.__input.read(count)

    ## read a line from the pipe
    def readline(self):
        return self.__input.readline()

    ## Do not block on reading
    #
    # Force reads from the specified file not to block
    #
    # This technique was taken from
    # http://mail.python.org/pipermail/python-dev/2005-March/052263.html
    # - that page suggests that this is the only portable way of
    # unblocking reads in Python
    def nonblocking_reads(self):
        fcntl.fcntl(self.__input.fileno(), fcntl.F_SETFL, fcntl.fcntl(self.__input.fileno(), fcntl.F_GETFL) | os.O_NONBLOCK)

    ## Block on reading
    #
    # Force reads from the specified file to block
    #
    # @sa nonblocking_reads()
    def blocking_reads(self):
        fcntl.fcntl(self.__input.fileno(), fcntl.F_SETFL, fcntl.fcntl(self.__input.fileno(), fcntl.F_GETFL) & ~os.O_NONBLOCK)


    ## Join this with another stream
    #
    # @param self this connection
    # @param conn2 connection to sew with
    # @param title to use in debugging messages
    # @param conn3 connection to monitor for an exit condition
    # @param verbose see below
    # @param conn1_name name to print for \c self in verbose mode
    # @param conn2_name name to print for \c conn2 in verbose mode
    # 
    # Pass two connections straight through to one another
    #
    # A lot of this program's work involves sewing two streams
    # together, making the output from stream 1 into the input for
    # stream 2 (and vice versa).  This function sews a pair of two-way
    # pipes together.
    #
    # If conn3 is specified, this function exits as soon as input is
    # seen from it.
    #
    # During debugging, it is sometimes useful to print the traffic
    # across a few streams.  Enabling this in normal code could slow
    # the assistant down, so to use this feature, you must set \c
    # verbose to \c True and edit the source of this function.
    def join(self, conn2, title, conn3=None, verbose=False, conn1_name="conn1", conn2_name="conn2"): #pylint: disable-msg=I0011,R0913

        if title is not None:
            debug.log('Joining', title)

        input_connections = [ self.fileno(), conn2.fileno() ]
        error_connections = [ self.fileno(), conn2.fileno() ]
        if conn3 is not None:
            input_connections.append(conn3.fileno())

        conn1_count = 0
        conn2_count = 0

        try:
            ewtd = False
            anything_actually_read = True
            while anything_actually_read and not ewtd:

                iwtd, dummy, ewtd = select.select(input_connections, [], error_connections)
                anything_actually_read = False
                #if verbose: debug.log(iwtd, ewtd)

                for connection in iwtd:
                    if self.fileno() == connection:
                        conn1_count = conn1_count+1
                        # Do not uncomment these in production versions of the assistant (see above)
                        #if verbose: debug.log(title, "- Reading from", conn1_name)
                        data = self.read()
                        #if verbose: debug.log(title, "- Read", len(data), "bytes from", conn1_name, " {" + data + "}")
                        anything_actually_read |= bool(data)
                        conn2.write(data)
                        #if verbose: debug.log(title, "- Written to", conn2_name)
                    else:
                        conn2_count = conn2_count+1
                        #if verbose: debug.log(title, "- Reading from", conn2_name)
                        data = conn2.read()
                        anything_actually_read |= bool(data)
                        #if verbose: debug.log(title, "- Read", len(data), "bytes from", conn2_name, " {" + data + "}")
                        self.write(data)
                        #if verbose: debug.log(title, "- Written to", conn1_name)

            debug.log("Finishing normally:\n\tewtd=", ewtd, "\n\tanything_actually_read=", anything_actually_read,
                      "\n\treads from", conn1_name + ":", conn1_count, "\n\treads from", conn2_name + ":", conn2_count)

        except (IOError, socket.error, select.error), err:
            debug.log("Finished erroneously:", err)
        finally:
            self.close()
            conn2.close()

        if title is not None:
            debug.log('Done', title)

        return 0


## Create a two-way pipe from standard input and output
class From_stdio(From_filehandles):
    ## Constructor
    #
    # @param self this connection
    # @param blocking_reads should the pipe block while reading?
    def __init__(self, blocking_reads=False):
        From_filehandles.__init__(self, sys.stdin, sys.stdout, blocking_reads)

## Create a two-way pipe from a pair of files, opening from_file first
class From_files_rw(From_filehandles):
    ## Constructor
    #
    # @param self this connection
    # @param from_file filename to read from
    # @param to_file filename to write to
    # @param blocking_reads should the pipe block while reading?
    #
    # When opening a pair of pipes, a deadlock occurs if both
    # processes try to open their reader first.
    def __init__(self, from_file, to_file, blocking_reads=False):
        from_fh = os.fdopen(os.open(from_file, os.O_RDONLY), 'rb')
        to_fh = open(to_file, 'ab')
        From_filehandles.__init__(self, from_fh, to_fh, blocking_reads)

## Create a two-way pipe from a pair of files, opening to_file first
class From_files_wr(From_filehandles):
    ## Constructor
    #
    # @param self this connection
    # @param from_file filename to read from
    # @param to_file filename to write to
    # @param blocking_reads should the pipe block while reading?
    #
    # When opening a pair of pipes, a deadlock occurs if both
    # processes try to open their reader first.
    def __init__(self, from_file, to_file, blocking_reads=False):
        to_fh = open(to_file, 'ab')
        from_fh = os.fdopen(os.open(from_file, os.O_RDONLY), 'rb')
        From_filehandles.__init__(self, from_fh, to_fh, blocking_reads)

## Create a two-way pipe from a command
#
# The command can either be a string or a list
class From_command(From_filehandles):

    ## Constructor
    #
    # @param self this connection
    # @param command command to run
    # @param blocking_reads should the pipe block while reading?
    # @param stderr where to send the command's standard error
    # @param killsig Signal with which to kill the process when the time comes
    #
    def __init__(self, command, blocking_reads=False, stderr=None, killsig=15):
        ## subprocess to read from and write to
        self.__pipe = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=stderr)
        ## Signal with which to kill the process when the time comes
        self.__killsig = killsig
        From_filehandles.__init__(self, self.__pipe.stdout, self.__pipe.stdin, blocking_reads)

    ## Close the pipe
    def close(self):
        self.__del__()
        From_filehandles.close(self)

    ## Destructor
    def __del__(self):
        try:
            os.kill(self.__pipe.pid, self.__killsig)

            if os.waitpid(self.__pipe.pid, os.WNOHANG)[0]: return
            else: time.sleep(1)

            if os.waitpid(self.__pipe.pid, os.WNOHANG)[0]: return
            else: time.sleep(1)

            if os.waitpid(self.__pipe.pid, os.WNOHANG)[0]: return

            os.kill(self.__pipe.pid, 9)

            self.__pipe.wait()
        except:
            pass

## Create a two-way pipe from a network (TCP) socket that has already connected
class From_connected_socket(From_filehandles):

    ## Constructor
    #
    # @param self this connection
    # @param sock socket to use
    def __init__(self, sock):
        From_filehandles.__init__(self, sock, sock, True)
        ## Socket that we communicate with
        self.__sock = sock

    ## Close the pipe
    def close(self):
        self.__sock.close()

    ## Write a string to the pipe
    def write(self, string):
        self.__sock.setblocking(1)
        self.__sock.sendall(string)

    ## Read up to 4096 bytes from the pipe
    def read(self, count=4096, blocking=0):
        timeout = self.__sock.gettimeout()
        self.__sock.setblocking(blocking)
        try:
            ret = self.__sock.recv(count)
        except socket.error:
            return ''
        self.__sock.settimeout(timeout)
        return ret

    ## Do not block on reading
    def nonblocking_reads(self):
        self.__sock.setblocking(0)

    ## Block on reading
    def blocking_reads(self):
        self.__sock.setblocking(1)

    ## Send data to a specific address
    def sendto(self, string, address):
        return self.__sock.sendto(string, address)

## Create a two-way pipe from a network (TCP) socket
class From_socket(From_connected_socket):

    ## Constructor
    #
    # @param self this connection
    # @param address address or socket (e.g. '127.0.0.1')
    # @param protocol networking protocol (e.g. \c socket.\c AF_INET)
    # @param timeout socket timeout
    #
    # When \c protocol is specified, \c address is an address in the
    # relevant format for the protocol.  When \c protocol is absent or
    # \c None, \c address must actually be a \c socket that's
    # listening for connections
    def __init__(self, address, protocol=None, timeout=1):

        # The one-argument version of the function takes a waiting socket instead of an address
        if protocol is None:
            sock, address = address.accept()

        # The two-argument version takes an address of the relevant type for this address family
        else:
            sock = socket.socket(protocol, socket.SOCK_STREAM)
            sock.settimeout(timeout)
            try:
                sock.connect(address)
                sock.settimeout(None)
                debug.log("Successfully connected to", address)
            except socket.error, err:
                debug.log("Failed to connect to", address, '-', err)
                raise

        From_connected_socket.__init__(self, sock)


## Create a two-way pipe from a network (UDP) socket
class From_udp_socket(From_connected_socket):

    ## Constructor
    def __init__(self, family, address):
        sock = socket.socket(family, socket.SOCK_DGRAM)
        sock.setblocking(1)
        ## Address to send data to
        self.__address = address
        From_connected_socket.__init__(self, sock)

    ## Write a string to the pipe
    def write(self, string):
        start = 0
        while start != len(string):
            start += self.sendto(string[start:], self.__address)
