#!/usr/bin/python

## @package unit_test-two_way_pipe
# Test the \c two_way_pipe class
# \author Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

## @file
# two-way pipe unit tests

## This file is distributed under the MIT license
LICENSE = """\
Copyright (c) 2008, Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import unittest, os, sys, re, time
import debugger, two_way_pipe, vnc_interaction, backend

# This program needs to be run on two computers at once, to test network issues.
# 
our_location, other_ip4, other_ip6 = sys.argv[1:]

## Test two_way_pipe.py
class TestTWP(unittest.TestCase):

    ## Test any type of pipe
    def pipe(self, pipe):
        for comment in ( "a", "b", """Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.""", "abcd" * 1024, "efgh" * 1025, "ijkl" * 4096):
            pipe.write(comment)

            # Strings should be split into 4096-byte chunks
            while len(comment) > 4096:
                self.assertEqual(pipe.read(), comment[:4096])
                comment = comment[4096:]

            self.assertEqual(pipe.read(), comment)

    ## Test two_way_pipe.From_filehandles
    def test_from_filehandles(self):
        # Test From_filehandles
        input_pipe, output_pipe = os.pipe()
        pipe = two_way_pipe.From_filehandles(os.fdopen(input_pipe, 'rb'), os.fdopen(output_pipe  , 'wb'))
        return self.pipe(pipe)

## Test vnc_interaction.py
class TestVNC(unittest.TestCase):

    ## Test any type of VNC server
    def vnc_server(self, server_type):
        server = vnc_interaction.get_server(True, server_type)

        signature = server.read(12, 1)

        #print "SIGNATURE: {%s}" % signature

        self.assert_(re.compile("RFB [0-9][0-9][0-9]\.[0-9][0-9][0-9]\n$").match(signature))

    ## Test the Vino VNC server
    def test_vino(self):

        import gconf
        gconf_client = gconf.client_get_default()

        entries = gconf_client.all_entries('/desktop/gnome/remote_access')

        self.vnc_server("vino")

        for entry in entries:
            self.assertEqual(gconf_client.get(entry.get_key()).to_string(), entry.get_value().to_string())

    ## Test the x11vnc VNC server
    def test_x11vnc(self): return self.vnc_server("x11vnc")
    ## Test the krfb VNC server
    def test_krfb  (self): return self.vnc_server("krfb")

## Test backend.py
class TestBackend(unittest.TestCase):

    ## Constructor
    def __init__(self):
        unittest.TestCase.__init__(self)
        self.data = None

    ## Test backend.friendly_fingerprint()
    def test_fingerprint(self):
        ## Test a single fingerprint
        def fingerprint_value(numeric_key, date_key):
            fingerprint_key = []
            for offset in range(120, -8, -8):
                fingerprint_key.append("%02x" % ( ( numeric_key & (0xFF << offset) ) >> offset))

            #print ":".join(fingerprint_key), date_key

            self.assertEquals(date_key, backend.friendly_fingerprint(":".join(fingerprint_key)))

        # Test that fingerprints are split up correctly into their respective dates
        fingerprint_value(0x00000000000000000000000000000000,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900")
        fingerprint_value(0x00000000000000000000000000000001,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:01 1900")
        fingerprint_value(0x00000000000000000000000100000000,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:01 1900\n" +
                          "Mon Jan  1 00:00:00 1900")

        fingerprint_value(0x00000000000000010000000000000000,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:01 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900")
        fingerprint_value(0x00000001000000000000000000000000,
                          "Mon Jan  1 00:00:01 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900")

        # From this point on, we assume that testing one date is as good as testing any other

        # One minute...
        fingerprint_value(0x0000000000000000000000000000003C,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:01:00 1900")

        # One day...
        fingerprint_value(0x00000000000000000000000000015180,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Tue Jan  2 00:00:00 1900")

        # One month...
        fingerprint_value(86400 * 31,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Thu Feb  1 00:00:00 1900")

        # No leap day in 1900...
        fingerprint_value(86400 * ( 31 + 28 ),
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Thu Mar  1 00:00:00 1900")

        # One year...
        fingerprint_value(86400 * 365 - 1,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Dec 31 23:59:59 1900")
        fingerprint_value(86400 * 365,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Tue Jan  1 00:00:00 1901")

        # No leap day in 1901...
        fingerprint_value(86400 * ( 365 + 31 + 28 ),
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Fri Mar  1 00:00:00 1901")

        # Two years...
        fingerprint_value(86400 * 365*2 - 1,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Tue Dec 31 23:59:59 1901")
        fingerprint_value(86400 * 365*2,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Wed Jan  1 00:00:00 1902")

        # No leap day in 1902...
        fingerprint_value(86400 * ( 365*2 + 31 + 28 ),
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Sat Mar  1 00:00:00 1902")

        # Three years...
        fingerprint_value(86400 * 365*3 - 1,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Wed Dec 31 23:59:59 1902")
        fingerprint_value(86400 * 365*3,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Thu Jan  1 00:00:00 1903")

        # No leap day in 1903...
        fingerprint_value(86400 * ( 365*3 + 31 + 28 ),
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Sun Mar  1 00:00:00 1903")


        # Four years...
        fingerprint_value(86400 * 365*4 - 1,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Thu Dec 31 23:59:59 1903")
        fingerprint_value(86400 * 365*4,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Fri Jan  1 00:00:00 1904")

        # Yes leap day in 1904!
        fingerprint_value(86400 * ( 365*4 + 31 + 28 ),
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Feb 29 00:00:00 1904")
        fingerprint_value(86400 * ( 365*4 + 31 + 29 ),
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Tue Mar  1 00:00:00 1904")

        # Five years...
        fingerprint_value(86400 * (365*5+1) - 1,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Sat Dec 31 23:59:59 1904")
        fingerprint_value(86400 * (365*5+1),
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Sun Jan  1 00:00:00 1905")

        # Maximum possible date
        fingerprint_value(0xFFFFFFFF,
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Mon Jan  1 00:00:00 1900\n" +
                          "Thu Feb  7 06:28:15 2036")

    ## Store callback data
    def generic_callback(self, data):
        self.data = data

    ## Test connections
    def test_connect(self):

        def do_test(other_ip):

            ## Back-end instance
            test_backend = backend.Backend()

            if 'local' == our_location:
                test_backend.connect((False, other_ip, False, False), self.generic_callback)
            else:
                test_backend.connect((True , ""      , False, False), self.generic_callback)

            self.data = None
            finished = False
            while test_backend.poll_message_queue() and not finished:
                if self.data is not None and self.data[0] == 1:
                    finished = True
                else:
                    time.sleep(1) # Wait until the back-end makes a connection

            self.assertEquals(len(self.data), 9)
            self.assertEquals(self.data[0], 1)

            if 'local' == our_location:
                self.assertEquals(self.data[6], backend.TRUST_SHARE)
            else:
                self.assertEquals(self.data[6], backend.TRUST_ACCESS)

            test_backend.start_session(self.data[6], self.generic_callback)

            self.data = self
            finished = False
            while test_backend.poll_message_queue() and not finished:
                if not isinstance(self.data, TestBackend):
                    finished = True
                else:
                    time.sleep(1) # Wait until the back-end makes a connection

            if 'local' == our_location:
                self.assertEquals(self.data, "RFB")
            else:
                debug.log(self.data)
                self.assert_(isinstance(self.data, tuple))
                self.assertEquals(self.data[0], "RFB")
                self.assert_(isinstance(self.data[1], str))

            test_backend.main_loop()

            time.sleep(1)

        do_test(other_ip4)
        do_test(other_ip6)

## Debugging and logging
debug = debugger.Debug(loglevel_modifier=-1)

if __name__ == '__main__':

    # Ensure the make process isn't killed if kill(0) is called
    os.setsid()

    debugger.kill_pids()
    try:
        os.remove(os.path.expanduser("~/unit_test_done"))
    except:
        pass

    if 'remote' == our_location:
        ## amount of output to print during unit tests
        verbosity = 0
    else:
        verbosity = debug.loglevel

    unittest.TextTestRunner(verbosity=verbosity).run(unittest.TestLoader().loadTestsFromTestCase(TestTWP))
    #unittest.TextTestRunner(verbosity=verbosity).run(unittest.TestLoader().loadTestsFromTestCase(TestVNC))
    unittest.TextTestRunner(verbosity=verbosity).run(unittest.TestLoader().loadTestsFromTestCase(TestBackend))

    open(os.path.expanduser("~/unit_test_done"), 'wb').write("Test complete")
    
