## @package vnc_interaction
# Connect to a VNC server or client
# \author Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>
#
# This package opens a connection with a VNC server or client, and
# returns a two-way pipe for communicating with it.

## @file
# Functions for interacting with VNC servers and clients

## This file is distributed under the MIT license
LICENSE = """\
Copyright (c) 2008, Andrew Sayers <andrew-remote-help-assistant@pileofstuff.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import os, sys, time, socket, traceback
import two_way_pipe, misc, debugger

## Debugging and logging
debug = debugger.Debug()

## VNC servers should run in quiet mode, even during debugging
QUIET_VNC_SERVER = True

## Create a two-way pipe from a vino server
class From_vino(two_way_pipe.From_socket): #pylint: disable-msg=I0011,R0903

    ## Constructor
    def __init__(self, address, protocol, vino_info):
        ## gconf values from before vino was activated
        self.__vino_info = None
        two_way_pipe.From_socket.__init__(self, (address, 5900), protocol)
        # Only use vino_info if the socket was successfully created
        self.__vino_info = vino_info

    ## Destructor
    def __del__(self):

        if self.__vino_info is not None:

            import gconf
            gconf_client = gconf.client_get_default()
            gconf_client.set('/desktop/gnome/remote_access/enabled', self.__vino_info[0])
            gconf_client.set('/desktop/gnome/remote_access/local_only', self.__vino_info[1])
            gconf_client.set('/desktop/gnome/remote_access/prompt_enabled', self.__vino_info[2])
            gconf_client.set('/desktop/gnome/remote_access/view_only', self.__vino_info[3])
            gconf_client.set('/desktop/gnome/remote_access/require_encryption', self.__vino_info[4])
            gconf_client.set('/desktop/gnome/remote_access/lock_screen_on_disconnect', self.__vino_info[5])
            gconf_client.set('/desktop/gnome/remote_access/authentication_methods', self.__vino_info[6])
            gconf_client.set('/desktop/gnome/remote_access/use_alternative_port', self.__vino_info[7])

## Create a two-way pipe from a VNC server
class From_vnc_server(two_way_pipe.From_socket): #pylint: disable-msg=I0011,R0903

    ## Give the server 2 seconds to respond to SIGKILL, then follow it up with a SIGTERM
    def __del__(self):

        debug.log("Killing VNC server ", self.__pid)

        os.kill(self.__pid, 15)

        if os.waitpid(self.__pid, os.WNOHANG)[0]: return
        for dummy in range(0, 2):
            time.sleep(1)
            if os.waitpid(self.__pid, os.WNOHANG)[0]: return

        os.kill(self.__pid, 9)
        os.waitpid(self.__pid, 0)

        debug.log("VNC server", self.__pid, "killed")

    ## Constructor
    def __init__(self, append_sockno, cmd):

        server, port = misc.listen_sockets('127.0.0.1', 5900, 5920)

        try:
            password = os.urandom(8)
        except:
            password = '12345' # Hey, it's good enough for my luggage...

        ## Process ID of the server
        self.__pid = os.fork()
        if self.__pid:
            two_way_pipe.From_socket.__init__(self, ('127.0.0.1', port), socket.AF_INET)
            self.write(password) #pylint: disable-msg=I0011,E1101

        else:
            debug.register_fork()
            sock = server.accept()[0]
            got_pass = sock.recv(len(password))
            if password != got_pass:
                debug.log("Bad password.  Exepcted: {"+password+"}, Got: {"+got_pass+"}")
                sys.exit(1)

            if append_sockno:
                cmd.append(str(sock.fileno()))
                os.execv(cmd[0], cmd)
            else:
                os.dup2(sock.fileno(), sys.stdout.fileno())
                os.dup2(sock.fileno(), sys.stdin.fileno())
                os.execv(cmd[0], cmd)

            debug.log("Impossible: exiting From_vnc_server.__init__()")
            sys.exit(1)



## Get a VNC or shell server
#
# @param allow_control share control of the desktop?
# @param force_server (DEBUGGING ONLY) force a specific server to be used
#
# This function tries to start the best server that's available, in
# the appropriate mode.  Servers are tried in this order:
#
# -# x11vnc
# -# krfb
# -# vino
#
# tightvncserver isn't used because it creates a new VNC session
# rather than sharing the current session.
#
# RDF (Terminal Services) servers aren't included at present, and
# may never be included, because it seems to use a more complex
# authentication protocol than could easily be supported by this
# program, and because there are no Linux RDP servers that share an
# existing X session rather than creating a new one.  See
# http://xrdp.sourceforge.net/documents/asession/index.html for a
# very brief discussion of RDP authentication.
#
# The first line sent by the server is used by the other side to
# auto-detect the session type.  At present, only VNC sessions are
# supported (beginning 'RFB NNN.NNN\\n').  Shell sessions used to be
# supported (beginning 'SHL 001.001\\n'), but were dropped because
# they would have required a lot of work to get right, and because
# there didn't seem to be any demand for them.
def get_server(allow_control, force_server=None):

    # First, try x11vnc.  This is preferred because it makes the least demands on the system (i.e. no KDE and no GNOME)
    if (force_server is None or "x11vnc" == force_server) and os.path.exists('/usr/bin/x11vnc'):

        debug.log("Trying x11vnc...")

        server_command = ['/usr/bin/x11vnc', '-nopw', '-once', '-inetd', '-nossl', '-safer', '-noncache']

        if QUIET_VNC_SERVER or debug.loglevel < 2:
            server_command.append('-q')

        if not allow_control:
            server_command.append('-viewonly')

        debug.log("Got x11vnc")

        return From_vnc_server(False, server_command)

    # Second, try to start a krfb session.  This is better than vino because it's easier to start and there's less lag for some reason.
    if (force_server is None or "krfb" == force_server) and os.path.exists('/usr/bin/krfb'):

        debug.log("Trying krfb...")

        krfbrc = open(misc.CONFIG_DIR + '/krfbrc', 'w')
        if allow_control:
            print >> krfbrc, "allowDesktopControl=true"
        else:
            print >> krfbrc, "allowDesktopControl=false"

        print >> krfbrc, """allowUninvited=true
confirmUninvitedConnection=false
disableBackground=false
disableXShm=false
enableSLP=true
preferredPort=-1
uninvitedPasswordCrypted=

[invitations]
invitation_num=0"""

        krfbrc.close()

        ret = From_vnc_server(True, ['/usr/bin/krfb', '--config', misc.CONFIG_DIR + '/krfbrc', '--kinetd'])
        os.remove(misc.CONFIG_DIR + '/krfbrc')

        debug.log("Got krfb")

        return ret

    if force_server is None or "vino" == force_server:
        # Try to start a vino session.  This is the final option because it can't be started on the command-line,
        # and (as of 12 July 2008) only supports IPv6
        try:
            debug.log("Trying vino...")

            import gconf
            gconf_client = gconf.client_get_default()

            # Note: I tried using a change set here, but Python segfaulted after reversing the changes
            vino_info = (
                gconf_client.get('/desktop/gnome/remote_access/enabled'),
                gconf_client.get('/desktop/gnome/remote_access/local_only'),
                gconf_client.get('/desktop/gnome/remote_access/prompt_enabled'),
                gconf_client.get('/desktop/gnome/remote_access/view_only'),
                gconf_client.get('/desktop/gnome/remote_access/require_encryption'),
                gconf_client.get('/desktop/gnome/remote_access/lock_screen_on_disconnect'),
                gconf_client.get('/desktop/gnome/remote_access/authentication_methods'),
                gconf_client.get('/desktop/gnome/remote_access/use_alternative_port')
                )

            n = gconf.Value('string')
            n.set_string('none')
            lst = gconf.Value('list')
            lst.set_list_type('string')
            lst.set_list([n])
            gconf_client.set_bool('/desktop/gnome/remote_access/local_only', True)
            gconf_client.set_bool('/desktop/gnome/remote_access/prompt_enabled', False)
            gconf_client.set_bool('/desktop/gnome/remote_access/view_only', not allow_control)
            gconf_client.set_bool('/desktop/gnome/remote_access/require_encryption', False)
            gconf_client.set_bool('/desktop/gnome/remote_access/lock_screen_on_disconnect', False)
            gconf_client.set     ('/desktop/gnome/remote_access/authentication_methods', lst)
            gconf_client.set_bool('/desktop/gnome/remote_access/enabled', True)
            gconf_client.set_bool('/desktop/gnome/remote_access/use_alternative_port', False)

            # vino has a bad habit of listening on IPv6 only, and can take a while to start up
            count = 0
            sock = None
            while sock is None and count < 4:
                count = count + 1
                try:
                    sock = From_vino('127.0.0.1', socket.AF_INET, vino_info)
                except socket.error:
                    try:
                        sock = From_vino('::1', socket.AF_INET6, vino_info)
                    except socket.error:
                        time.sleep(1)

            if sock:
                # In case anybody tries to ride in on our coat-tails
                # Note: at the time of writing, this doesn't work.
                # Hopefully, future vinos will use the prompt_enabled
                # setting from when the connection was established.
                gconf_client.set_bool('/desktop/gnome/remote_access/prompt_enabled', True)

                debug.log("Got vino")

                return sock
            else:
                debug.log("Failed to start vino:", sys.exc_info()[0])
                gconf_client.set('/desktop/gnome/remote_access/enabled', vino_info[0])
                gconf_client.set('/desktop/gnome/remote_access/local_only', vino_info[1])
                gconf_client.set('/desktop/gnome/remote_access/prompt_enabled', vino_info[2])
                gconf_client.set('/desktop/gnome/remote_access/view_only', vino_info[3])
                gconf_client.set('/desktop/gnome/remote_access/require_encryption', vino_info[4])
                gconf_client.set('/desktop/gnome/remote_access/lock_screen_on_disconnect', vino_info[5])
                gconf_client.set('/desktop/gnome/remote_access/authentication_methods', vino_info[6])
                gconf_client.set('/desktop/gnome/remote_access/use_alternative_port', vino_info[7])

        except IOError:
            debug.log("Failed to start vino:", sys.exc_info()[0])

    return None

## Get a VNC or shell client
#
# This function is the counterpart to get_server().
def get_client():
    try:
        sock = None

        while sock is None:
            sock, port = misc.listen_sockets('127.0.0.1', 5900, 5920)

        return sock, str(port)

    except:
        debug.log("Unexpected error, stack trace follows")
        traceback.print_exc(file=sys.stderr)
